# utils.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import numba
import torch

from tqdm import tqdm


def _validate_input(X, name, shape=None, dtype=None, min_value=None,
	max_value=None, ohe=False, ohe_dim=1, allow_N=False):
	"""An internal function for validating properties of the input.

	This function will take in an object and verify characteristics of it, such
	as the type, the datatype of the elements, its shape, etc. If any of these
	characteristics are not met, an error will be raised.


	Parameters
	----------
	X: torch.Tensor
		The object to be verified.

	name: str
		The name to reference the tensor by if an error is raised.

	shape: tuple or None, optional
		The shape the tensor must have. If a -1 is provided at any axis, that
		position is ignored.  If not provided, no check is performed. Default is
		None.

	dtype: torch.dtype or None, optional
		The dtype the tensor must have. If not provided, no check is performed.
		Default is None.

	min_value: float or None, optional
		The minimum value that can be in the tensor, inclusive. If None, no
		check is performed. Default is None.

	max_value: float or None, optional
		The maximum value that can be in the tensor, inclusive. If None, no
		check is performed. Default is None.

	ohe: bool, optional
		Whether the input must be a one-hot encoding, i.e., only consist of
		zeroes and ones. Default is False.

	allow_N: bool, optional
		Whether to allow the return of the character 'N' in the sequence, i.e.
		if pwm at a position is all 0's return N. Default is False.


	Returns
	X: torch.Tensor
		The same object, unmodified, for convenience.
	"""

	if not isinstance(X, torch.Tensor):
		raise ValueError("{} must be a torch.Tensor object".format(name))

	if shape is not None:
		if len(shape) != len(X.shape):
			raise ValueError("{} must have shape {}".format(name, shape))

		for i in range(len(shape)):
			if shape[i] != -1 and shape[i] != X.shape[i]:
				raise ValueError("{} must have shape {}".format(name, shape))


	if dtype is not None and X.dtype != dtype:
		raise ValueError("{} must have dtype {}".format(name, dtype))

	if min_value is not None and X.min() < min_value:
		raise ValueError("{} cannot have a value below {}".format(name,
			min_value))

	if max_value is not None and X.max() > max_value:
		raise ValueError("{} cannot have a value above {}".format(name,
			max_value))

	if ohe:
		values = torch.unique(X)
		if len(values) != 2:
			raise ValueError("{} must be one-hot encoded.".format(name))

		if not all(values == torch.tensor([0, 1], device=X.device)):
			raise ValueError("{} must be one-hot encoded.".format(name))

		if allow_N:
			if not torch.all(torch.sum(X, axis=ohe_dim) <= 1):
				raise ValueError("{} must be one-hot encoded.".format(name) +
					"and contain unknown characters as all-zeroes.")
		else:
			if not torch.all(X.sum(axis=ohe_dim) == 1):
				raise ValueError("{} must be one-hot encoded ".format(name) +
					"and cannot have unknown characters.")				

	return X


def _cast_as_tensor(value, dtype=None):
	"""Cast your input as a torch tensor.

	This function will take some array-like input and cast it as a torch
	tensor with optionally a desired dtype. If X is already a torch datatype,
	ensure that the dtype matches.


	Parameters
	----------
	value: array-like
		An array-like object to be cast into a torch.tensor

	dtype: torch.dtype
		A torch dtype to cast the values in the torch.tensor to


	Returns
	-------
	value: torch.tensor
		A tensor that has been created from the array-like with the provided
		dtype.
	""" 

	if value is None:
		return None

	_tdtype = (torch.nn.Parameter, torch.Tensor, torch.masked.MaskedTensor)
	if isinstance(value, _tdtype):
		if dtype is None:
			return value
		elif value.dtype == dtype:
			return value
		else:
			return value.type(dtype)
			
	if isinstance(value, list):
		if all(isinstance(v, numpy.ndarray) for v in value):
			value = numpy.array(value)
		
	if isinstance(value, (float, int, list, tuple, numpy.ndarray)):
		if dtype is None:
			return torch.tensor(value)
		else:
			return torch.tensor(value, dtype=dtype)


def example_to_fasta_coords(example_df, loci_df, window=None, one_indexed=False):
	"""Converts coordinates within a given example to those in a FASTA file.
	
	Many analyses involve extracting windows from the genome and processing them
	independently. When coordinates are identified on these windows, e.g.,
	seqlets or other spans, all that is recorded is the example index and the
	relative positions from the start of the extracted window. Frequently, though,
	one wants to know these coordinates on the original genome or other entity
	encoded in a FASTA file. This function cross-references the example indexes
	and BED-formatted locus file that the examples were extracted from to get
	the coordinates on the original sequence.
	
	This cross-referencing has several steps.
	
		(1) For each span in `example_df` pull out the chrom, start, and end
		positions in `loci_df`. If `window` is provided, do not use the start and
		end in the file but, rather, extract a window centered on the middle of the
		provided coordinates.
		
		(2) Use the chrom directly, and add the start and end coordinates to the
		start of the extracted window from step (1).
	
	This function assumes by default that `example_df` is zero-indexed, in the 
	sense that the minimum start is 0. When doing the conversion, the start and
	end are simply added to the value in `loci_df`. This means that `loci_df` can
	be either zero- or one-indexed and the resulting df will match this indexing.
	If `example_df` is one-indexed, use the optional parameter.
	
	
	Parameters
	----------
	example_df: pandas.DataFrame
		A BED-formatted dataframe where the first three columns must be the example
		index, the start, and the end coordinates. It does not matter what these
		columns are named, and there can be additional columns past these.
	
	loci_df: pandas.DataFrame
		A BED-formatted dataframe where the first three columns must be the example
		index, the start, and the end coordinates. It does not matter what these
		columns are named, and there can be additional columns past these.
		
		NOTE: some processing steps may skip over elements in BED files and return
		a tensor with fewer elements than there are entries in the original BED
		file. Please check to make sure this has not happened, or everything will
		be misaligned.
	
	window: int or None, optional
		If None, use the start and end coordinates in `loci_df` directly. If an
		integer, these start and end coordinates will be replaced with a window
		centered on the middle of these original coordinates that spans this
		window length. This is useful when processing methods extracted windows
		of uniform length from a peak file, regardless of the length of that
		peak. Default is None.
		
	one_indexed: bool, optional
		Whether `example_df` is one-indexed. If it were one-indexed, the smallest
		start value would be 1 and so 1 would have to be subtracted from each
		correction. If False, assumes that the smallest start value is 0 and no
		change needs to be made. Default is False.
		
	
	Returns
	-------
	coords_df: pandas.DataFrame
		A dataframe of corrected coordinates where the first three columns are
		chrom, start, and end, and correspond to the same coordinate system as
		in `loci_df`. Put another way, if `loci_df` contains peaks on a reference
		genome, this returned file will now contain coordinates on that reference.
		The names of the first three columns will be replaced with the names of the
		first three columns in `loci_df`, and all other columns in the file will
		be left unchanged.
	"""
	
	loci_chroms = loci_df.iloc[:, 0].values
	loci_starts = loci_df.iloc[:, 1].values
	loci_ends = loci_df.iloc[:, 2].values
	
	if window is not None:
		mid = (loci_ends + loci_starts) // 2
		
		loci_starts = mid - window // 2
		loci_ends = mid + window // 2 + window % 2
	
	coords_ = example_df.iloc[:, :3].values
	idxs = coords_[:, 0]
	
	example_chroms = loci_chroms[idxs]
	example_starts = loci_starts[idxs] + coords_[:, 1] - one_indexed
	example_ends = loci_starts[idxs] + coords_[:, 2] - one_indexed
	
	names = loci_df.columns[:3]
	key_map = dict(zip(example_df.columns[:3], names))
	coords_df = example_df.copy()
	coords_df = coords_df.rename(key_map, axis='columns')
	coords_df[names[0]] = example_chroms
	coords_df[names[1]] = example_starts
	coords_df[names[2]] = example_ends
	return coords_df


def characters(pwm, alphabet=['A', 'C', 'G', 'T'], force=False, allow_N=False):
	"""Converts a PWM/one-hot encoding to a string sequence.

	This function takes in a PWM or one-hot encoding and converts it to the
	most likely sequence. When the input is a one-hot encoding, this is the
	opposite of the `one_hot_encoding` function.


	Parameters
	----------
	pwm: torch.tensor, shape=(len(alphabet), seq_len)
		A numeric representation of the sequence. This can be one-hot encoded
		or contain numeric values. These numerics can be probabilities but can
		also be frequencies.

	alphabet : set or tuple or list
		A pre-defined alphabet where the ordering of the symbols is the same
		as the index into the returned tensor. This is used to determine the
		letters in the returned sequence. Default is the DNA alphabet.

	force: bool, optional
		Whether to force a sequence to be produced even when there are ties.
		At each position that there is a tight, the character earlier in the
		sequence will be used. Default is False.
  
	allow_N: bool, optional
		Whether to allow the return of the character 'N' in the sequence, i.e.
		if pwm at a position is all 0's return N. Default is False.


	Returns
	-------
	seq: str
		A string where the length is the second dimension of PWM.
	"""
 
	#if (batch, alphabet_size, motif_size) and batch = 1, remove batch axis
	if len(pwm.shape) == 3 and pwm.shape[0] == 1:
		pwm = pwm[0]

	if len(pwm.shape) != 2:
		raise ValueError("PWM must have two dimensions where the " +
			"first dimension is the length of the alphabet and the second " +
			"dimension is the length of the sequence.")

	if pwm.shape[0] != len(alphabet):
		raise ValueError("PWM must have the same alphabet size as the " +
			"provided alphabet.")

	pwm_ismax = pwm == pwm.max(dim=0, keepdims=True).values
	if pwm_ismax.sum(axis=0).max() > 1 and force == False and allow_N == False:
		raise ValueError("At least one position in the PWM has multiple " +
			"letters with the same probability.")

	alphabet = numpy.array(alphabet)
	if isinstance(pwm, torch.Tensor):
		pwm = pwm.numpy(force=True)

	if allow_N:
		n_inds = numpy.where(pwm.sum(axis=0)==0)[0]
		dna_chars = alphabet[pwm.argmax(axis=0)]
		dna_chars[n_inds] = 'N'
	else:
		dna_chars = alphabet[pwm.argmax(axis=0)]
	
	return ''.join(dna_chars)


@numba.njit("void(int8[:, :], int8[:], int8[:])", cache=True)
def _fast_one_hot_encode(X_ohe, seq, mapping):
	"""An internal function for quickly converting bytes to one-hot indexes."""

	for i in range(len(seq)):
		idx = mapping[seq[i]]
		if idx == -1:
			continue

		if idx == -2:
			raise ValueError("Encountered character that is not in " + 
				"`alphabet` or in `ignore`.")
			
		X_ohe[i, idx] = 1


def one_hot_encode(sequence, alphabet=['A', 'C', 'G', 'T'], dtype=torch.int8, 
	ignore=['N'], desc=None, verbose=False, **kwargs):
	"""Converts a string or list of characters into a one-hot encoding.

	This function will take in either a string or a list and convert it into a
	one-hot encoding. If the input is a string, each character is assumed to be
	a different symbol, e.g. 'ACGT' is assumed to be a sequence of four 
	characters. If the input is a list, the elements can be any size.

	Although this function will be used here primarily to convert nucleotide
	sequences into one-hot encoding with an alphabet of size 4, in principle
	this function can be used for any types of sequences.

	Parameters
	----------
	sequence : str or list
		The sequence to convert to a one-hot encoding.

	alphabet : set or tuple or list
		A pre-defined alphabet where the ordering of the symbols is the same
		as the index into the returned tensor, i.e., for the alphabet ['A', 'B']
		the returned tensor will have a 1 at index 0 if the character was 'A'.
		Characters outside the alphabet are ignored and none of the indexes are
		set to 1. Default is ['A', 'C', 'G', 'T'].

	dtype : str or torch.dtype, optional
		The data type of the returned encoding. Default is int8.

	ignore: list, optional
		A list of characters to ignore in the sequence, meaning that no bits
		are set to 1 in the returned one-hot encoding. Put another way, the
		sum across characters is equal to 1 for all positions except those
		where the original sequence is in this list. Default is ['N'].


	Returns
	-------
	ohe : numpy.ndarray
		A binary matrix of shape (alphabet_size, sequence_length) where
		alphabet_size is the number of unique elements in the sequence and
		sequence_length is the length of the input sequence.
	"""

	for char in ignore:
		if char in alphabet:
			raise ValueError("Character {} in the alphabet ".format(char) + 
				"and also in the list of ignored characters.")

	if isinstance(alphabet, list):
		alphabet = ''.join(alphabet)

	ignore = ''.join(ignore)

	e = "utf8"
	seq_idxs = numpy.frombuffer(bytearray(sequence, e), dtype=numpy.int8)
	alpha_idxs = numpy.frombuffer(bytearray(alphabet, e), dtype=numpy.int8)
	ignore_idxs = numpy.frombuffer(bytearray(ignore, e), dtype=numpy.int8)

	one_hot_mapping = numpy.zeros(256, dtype=numpy.int8) - 2
	for i, idx in enumerate(alpha_idxs):
		one_hot_mapping[idx] = i

	for i, idx in enumerate(ignore_idxs):
		one_hot_mapping[idx] = -1


	n, m = len(sequence), len(alphabet)

	one_hot_encoding = numpy.zeros((n, m), dtype=numpy.int8)
	_fast_one_hot_encode(one_hot_encoding, seq_idxs, one_hot_mapping)
	return torch.from_numpy(one_hot_encoding).type(dtype).T


def reverse_complement(seq, complement_map={"A": "T", "C": "G", "G": "C", 
	"T": "A"}, allow_N=True):
	"""Return the reverse complement of a single sequence.

	This function will take in a single one-hot encoding of a sequence, or a
	single string, and return the reverse complement. If the input is a torch
	tensor, the encoding is simply flipped along both axes. If the input is
	a string, it is flipped and then each value is flipped according to the
	provided complement map. 

	Note that this function will not convert your sequence to upper-case or
	modify it in any other manner.

	
	Parameters
	----------
	seq: str or torch.Tensor w/ shape (alphabet_size, length)
		The sequence to be reverse complemented.
	
	complement_map: dict, optional
		The ordering and complement of each nucleotide. When the input is a
		string, this is used to directly convert characters. When the input is
		a torch tensor, the ordering of *keys* is assumed to be the order of 
		characters in the alphabet and the manner in which to flip depends on 
		it. Default is the nucleotide alphabet.

	allow_N: bool, optional
		Whether to allow N characters when doing the reverse complement. Only
		matters when reverse complementing strings. Default is True.
	
	Returns
	-------
	rev_comp: str or torch.Tensor w/ shape (alphabet_size, length)
		The reverse complemented string or 
	"""

	if isinstance(seq, str):
		seq_rc = []

		for char in seq:
			if char in complement_map:
				seq_rc.append(complement_map[char])
			elif char == 'N' and allow_N:
				seq_rc.append('N')
			else:
				raise ValueError("'{}' not in complement map".format(char))

		seq_rc = ''.join(reversed(seq_rc))

	elif isinstance(seq, torch.Tensor):

		chars = list(complement_map.keys())
		idxs = [chars.index(char) for char in complement_map.values()]

		seq_rc = torch.flip(seq, dims=(-1,))[idxs]

	return seq_rc


def random_one_hot(shape, probs=None, dtype='int8', random_state=None):
	"""Generate random one-hot encodings. Useful for debugging.

	This function will generate random one-hot encodings where the second to
	last dimension has a single element being a one and every other element
	being a zero. Primary used for debugging. 


	Parameters
	----------
	shape: tuple
		The shape of the 3D tensor to generate.

	probs: tuple, list, numpy.ndarray, or None optional
		A 2D array of probabilities where the first dimension is the batch size
		equal to the batch size of `X` and the second dimension is the alphabet
		size. The values should be the probability of that character occuring
		in that sequence. If a batch size of 1 is used when the batch size of
		X is greater than 1, the same probabilities are used for each sequence.
		The sum of probabilities across the alphabet axis must be equal to 1.
		If None, use a uniform distribution across the axis specified in shape.
		Default is [[0.25, 0.25, 0.25, 0.25]].

	dtype: str or numpy.dtype, optional
		The datatype to return the matrix as. Default is 'int8'.

	random_state: int or numpy.random.RandomState or None, optional
		The random state to use for generation. If None, do not use a 
		deterministic seed. Default is None.


	Returns
	-------
	ohe: torch.Tensor
		A tensor with the specified shape that is one-hot encoded.
	"""

	if not isinstance(shape, tuple) or len(shape) != 3:
		raise ValueError("Shape must be a tuple with 3 dimensions.")

	if not isinstance(random_state, numpy.random.RandomState):
		random_state = numpy.random.RandomState(random_state)

	if isinstance(probs, list):
		probs = numpy.array(probs)
		
	n = shape[1]
	ohe = numpy.zeros(shape, dtype=dtype)

	for i in range(ohe.shape[0]):
		if probs is None:
			probs_ = None
		elif probs.ndim == 1:
			probs_ = probs
		elif probs.shape[0] == 1:
			probs_ = probs[0]
		else:
			probs_ = probs[i] 

		choices = random_state.choice(n, size=shape[2], p=probs_)
		ohe[i, choices, numpy.arange(shape[2])] = 1 

	return torch.from_numpy(ohe)


def chunk(X, size=1024, overlap=0):
	"""Chunk a set of sequences into overlapping blocks.

	This function will take a set of sequences of variable length and will
	return a set of fixed-length blocks that tile the sequence with a fixed
	amount of overlap. This is useful when applying a method, such as FIMO or
	even a larger predictive model, to variable length sequences such as
	chromosomes.

	Unless the total sequence length is a multiple of the size, the final chunk
	produced from a sequence will be shorter than the size. In this case, the
	final chunk is excluded because it would not be an accurate representation
	of that sequence regardless.


	Parameters
	----------
	X: list of torch.Tensors
		A list of one-hot encoded sequences that each are of shape (4, -1).

	size: int, optional
		The size of the chunks to produce from the sequence. Default is 1024.

	overlap: int, optional
		The overlap between adjacent chunks. This is, essentially, the stride
		of the unrolling process. Default is 0.


	Returns
	-------
	y: torch.Tensor, shape=(-1, len(alphabet), size)
		A single tensor containing strides across the sequences. The chunks
		are concatenated together across examples such that the final chunk
		from the first item is followed by first chunk from the second item.
	"""

	if not isinstance(X, list):
		raise ValueError("X must be a list of tensors.")

	if not isinstance(size, int) or size <= 0:
		raise ValueError("size must be a positive integer.")

	if not isinstance(overlap, int) or overlap < 0:
		raise ValueError("overlap must be a non-negative integer.")

	return torch.cat([x.unfold(-1, size, size-overlap).permute(1, 0, 2) 
		for x in X], dim=0)


def unchunk(X, lengths=None, overlap=0):
	"""Unchunk fixed-length segments back into variable-length sequences.

	After chunking a set of variable length sequences into fixed-length chunks
	and applying some method on the chunks that results in some bp-resolution
	result, merge the chunks back into a tensor of the same length as the
	original sequence. When the final chunk was discarded due to the original
	sequence not being divisible by the chosen size, that portion will not
	be reconstructed by this method. 

	The overlap value should be the same as when the sequence was chunked, and
	should correspond to the number of positions that are shared across adjacent
	examples. When overlap is set to a value greater than 0, half of the overlap
	goes to the elements as follows:

		<------------*    |
				  overlap 
				|    *-----------> 


	Parameters
	----------
	X: list or numpy.ndarray or torch.tensor, shape=(-1, n_outputs, size)
		A set of fixed-length tensors for any number of outputs. Usually the
		result of applying `chunk` and then some form of model.

	lengths: list or numpy.ndarray or torch.tensor, shape=(-1,) or None
		The *original lengths* of the elements that were chunked. This is not
		the number of chunks produced, which will be influenced by the overlap,
		but the actual length in bp of the original sequences. If None, assume
		that all chunks in `X` come from a single variable-length sequence.
		Default is None.

	overlap: int, optional
		The number of bp overlap between adjacent chunks. Default is 0.


	Returns
	-------
	y: list of torch.Tensors or one torch.Tensor, shape=(n_outputs, -1)
		A list of variable-length tensors if `lengths` is provided, otherwise
		a single tensor.
	"""

	X = _cast_as_tensor(X)
	if X.ndim <= 2:
		raise ValueError("`X` must have at least three dimensions, with the "
			"last dimension corresponding to length.")

	lengths = _validate_input(_cast_as_tensor(lengths), "lengths", shape=(-1,), 
		min_value=0)

	size = X.shape[-1]
	lengths = (lengths - size) // (size - overlap) + 1
	lengths_csum = 0

	y = []
	for length in lengths:
		X_ = X[lengths_csum:lengths_csum+length]

		if overlap > 0:
			s = overlap // 2
			e = -(overlap - s)

			if X_.shape[0] == 1:
				X_ = X_[..., s:e].moveaxis(0, -2).reshape(*X_.shape[1:-1], -1)
			elif X_.shape[0] == 2:
				X_ = torch.cat([X_[0, ..., :e], X_[1, ..., s:]], dim=-1)
			else:
				X_ = torch.cat([X_[0, ..., :e],
								X_[1:-1, ..., s:e].moveaxis(0, -2).reshape(*X_.shape[1:-1], -1),
								X_[-1, ..., s:]], dim=-1)
		else:
			X_ = X_.moveaxis(0, -2).reshape(*X_.shape[1:-1], -1)

		y.append(X_)
		lengths_csum += length

	return y
	

def pwm_consensus(X):
	"""Take in a PWM and return the consensus.

	This function will take in a PWM that encodes the probabilities of each
	character at each position and will return the consensus, which is the
	most likely individual sequence according to that PWM. This is done by
	taking the argmax at each position. When multiple characters at the same
	position have the maximum probability, the character in the earlier
	numerical position is chosen. If a column has a sum equal to 0 (no
	characters allowed there) the returned consensus also has entirely 0s.


	Parameters
	----------
	X: torch.Tensor, numpy.ndarray, shape=(alphabet_len, pwm_len)
		A PWM containing the probabilities of each character at each position.


	Returns
	-------
	Y: torch.Tensor, shape=(alphabet_len, pwm_len)
		A PWM of the same shape as `X` except with the maximum-value character
		set to 1.
	"""

	X = _cast_as_tensor(X)
	_validate_input(X, "X", shape=(-1, -1), min_value=0, max_value=1)

	alpha_idxs = X.argmax(dim=0)

	Y = torch.zeros_like(X)
	Y[alpha_idxs, torch.arange(X.shape[-1])] = 1
	Y[:, X.sum(dim=0) == 0] = 0
	return Y


def extract_signal(loci, X, verbose=False):
	"""Extracts the signal at coordinates from a tensor of examples.

	This function takes in a dataframe with the first three columns being the
	example index, the start (inclusive) and the end (not inclusive) and
	returns the signal sum across those coordinates from the tensor. This can
	be used, for instance, to extract attributions or genomics signal from
	windows. This sum is done separately for each signal in the tensor.


	Parameters
	----------
	loci: pandas.DataFrame
		A set of loci to extract signal from. Multiple loci can be present on
		the same example in X and the starts and ends can overlap.

	X: torch.Tensor, numpy.ndarray, shape=(-1, n_signals, sequence_length)
		A 3D tensor where the first dimension corresponds to the examples, the
		second dimension corresponds to the number of signals being measured
		at each position (e.g., different ChIP-seq tracks), and the third
		position corresponds to the sequence length.

	verbose: bool, optional
		Whether to print a progress bar tracking the extraction process. Default
		is False.


	Returns
	-------
	Y: torch.Tensor, shape=(n_loci, n_signals)
		The sum of the signal across all positions for each of the loci. The
		returned value is the sum of this signal.
	"""

	_validate_input(X, "X", shape=(-1, -1, -1))

	Y = torch.zeros(loci.shape[0], X.shape[1], dtype=X.dtype, device=X.device)

	loci = loci.values[:, :3].astype(int)
	for i, (idx, start, end) in enumerate(tqdm(loci, disable=not verbose)):
		Y[i] = X[idx, :, start:end].sum(dim=-1)

	return Y



# variant.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import torch
import itertools

from .io import extract_loci

from .utils import one_hot_encode
from .utils import _cast_as_tensor

from .ersatz import delete
from .ersatz import insert

from .predict import predict
from .marginalize import marginalize


def substitution_effect(model, X, substitutions, args=None, func=predict, 
	additional_func_kwargs=None, **kwargs):
	"""Apply a function before and after including one or more substitutions.

	This function will calculate the effect that substitutions have on the
	output from a model. Any number of substitutions can be added to each
	sequence and the provided `func` is applied before and after these
	substitutions are included in the sequences. By default, this `func` is the
	prediction function and so the results are the difference in predictions
	before and after substitutions are made, but if `func` is something else, 
	such as `deep_lift_shap`, the results will be the attributions before and
	after substitutions are made. At least one substitution should be provided 
	per sequence for the results to be different.

	The substitutions provided must be individual variants, i.e., that each row 
	in the tensor corresponds to a single substitution in a single example, but
	one can encode longer variants (e.g., entire motifs or just multiple
	characters) by passing in multiple rows with adjacent positions. 

	Note that substitutions are not insertions. A substitution involves changing
	one character to another character. An insertion involves adding a new
	character into a sequence and requires trimming the edges afterward. Indels
	can be represented as substitutions as long as the indel is one insertion
	and also one deletion.


	Parameters
	----------
	model: torch.nn.Module
		A PyTorch model to use for making predictions. These models can take in
		any number of inputs and make any number of outputs. The additional
		inputs must be specified in the `args` parameter.

	X: torch.tensor, shape=(-1, len(alphabet), length)
		A one-hot encoded set of sequences to have substitutions included in.

	substitutions: torch.tensor, shape=(-1, 3)
		A set of variants that should be substituted into each sequence. This
		tensor is formatted like a COO-sparse matrix where each row is a single
		variant, the first column is the index in `X`, the second index is the
		position in that example, and the third index is the index into the
		alphabet that should be present at that position (overriding whatever
		is currently there). 

	args: tuple or None, optional
		An optional set of additional arguments to pass into the model. If
		provided, each element in the tuple or list is one input to the model
		and the element must be formatted to be the same batch size as `X`. If
		None, no additional arguments are passed into the forward function. This
		argument is provided here because the args must be copied for each
		shuffle that occurs. Default is None.

	func: function, optional
		A function to apply before and after incorporating the substitution. 
		Default is `predict`.

	additional_func_kwargs: dict, optional
		Additional named arguments to pass into the function when it is called.
		This is provided as an alternate path to route arguments into the 
		function in case they overlap, name-wise, with those in this function,
		or if you want to be absolutely sure that the arguments are making
		their way into the function. Default is {}.

	kwargs: optional
		Additional named arguments that will get passed into the function when
		it is called. Default is no arguments are passed in.


	Returns
	-------
	y_before: torch.Tensor
		The output from `func` before variants are included.

	y_after: torch.Tensor
		The output from `func` after the variants are included.
	"""

	substitutions = _cast_as_tensor(substitutions)

	additional_func_kwargs = additional_func_kwargs or {}

	X_var = torch.clone(X)
	X_var[substitutions[:, 0], :, substitutions[:, 1]] = 0
	X_var[substitutions[:, 0], substitutions[:, 2], substitutions[:, 1]] = 1

	y_before = func(model, X, args=args, **additional_func_kwargs, **kwargs)
	y_after = func(model, X_var, args=args, **additional_func_kwargs, **kwargs)
	return y_before, y_after


def deletion_effect(model, X, deletions, left=False, args=None, func=predict,
	additional_func_kwargs=None, **kwargs):
	"""Apply a function before and after deleting characters from a sequence.

	This function will calculate the effect that insertions have on the
	output from a model. Any number of deletions can be specified for each
	sequence and the provided `func` is applied before and after the deletions
	are taken into account. By default, this `func` is the prediction function
	and so the results are the difference in predictions before and after the
	deletions are made, but if `func` is something else, such as 
	`deep_lift_shap`, the results will be the attributions before and after the
	deletions are made. At least one deletion should be provided per sequence
	for the results to be different.

	The deletions provided must be individual characters, i.e., that each row
	in the tensor corresponds to a single deletion in a single example, but one
	can encode longer deletions (e.g., entire motifs or just multiple 
	characters) by passing in multiple rows with adjacent positions.

	Importantly, because models assume a fixed input window but deletons are by
	definition changing the length of the sequence, the provided sequences must
	be of length `model_length + max_deletions_per_sequence`. Basically, if the
	maximum number of deletions in a sequence is equal to 12 and the model
	expects a tensor of length 100 then every sequence provided must be of
	length 112, even if there are fewer than 12 deletions in a particular
	sequence.

	Simply removing all of the specified characters will lead to a set of
	sequences of differing lengths if there are a different number of deletions
	in each sequence. In order to make these sequences all the same length, we
	need to trim from each sequence a number of positions such that in total
	(these additional bases + the number of deletions provided by the user)
	characters removed per example is the same for every example. Because there
	are two ways we can trim positions -- either starting from the left end of 
	the sequence or from the right end of it -- you can use the `left` parameter
	to specify that you should trim positions on the left (`left=True`) or from
	the right (`left=False`, the default) of the sequence. 


	Parameters
	----------
	model: torch.nn.Module
		A PyTorch model to use for making predictions. These models can take in
		any number of inputs and make any number of outputs. The additional
		inputs must be specified in the `args` parameter.

	X: torch.tensor, shape=(-1, len(alphabet), length + max_deletions)
		A one-hot encoded set of sequences to have deletions included in.

	deletions: torch.tensor, shape=(-1, 2)
		A set of deletions indicating characters that should be removed from
		each sequence. Each row should be a single deletion with the first
		column corresponding to the index and the second column corresponding
		to the position within that index. Multiple deletions can occur in each
		example. 

	left: bool, optional
		If False, use the first `n` positions to run through the model before
		making the deletion where `n` is the expected tensor length. If True,
		use the last `n` positions. Basically, whether we trim positions from
		the left or the right when getting sequences of the same length.
		Default is False.

	args: tuple or None, optional
		An optional set of additional arguments to pass into the model. If
		provided, each element in the tuple or list is one input to the model
		and the element must be formatted to be the same batch size as `X`. If
		None, no additional arguments are passed into the forward function. This
		argument is provided here because the args must be copied for each
		shuffle that occurs. Default is None.

	func: function, optional
		A function to apply before and after incorporating the substitution. 
		Default is `predict`.

	additional_func_kwargs: dict, optional
		Additional named arguments to pass into the function when it is called.
		This is provided as an alternate path to route arguments into the 
		function in case they overlap, name-wise, with those in this function,
		or if you want to be absolutely sure that the arguments are making
		their way into the function. Default is {}.

	kwargs: optional
		Additional named arguments that will get passed into the function when
		it is called. Default is no arguments are passed in.


	Returns
	-------
	y_before: torch.Tensor
		The output from `func` before variants are included.

	y_after: torch.Tensor
		The output from `func` after the variants are included.
	"""

	deletions = _cast_as_tensor(deletions)

	additional_func_kwargs = additional_func_kwargs or {}

	mask = torch.zeros_like(X[:, 0]).type(torch.int32)
	mask[deletions[:, 0], deletions[:, 1]] = 1
	
	counts = mask.sum(dim=-1)
	counts = abs(counts - counts.max())

	m = mask if left == True else torch.flip(mask, dims=(-1,))
	flank = torch.cumsum(1 - m, dim=-1) <= counts[:, None]
	mask += (flank if left == True else torch.flip(flank, dims=(-1,)))
	mask = (1 - mask).type(torch.bool)
	mask = mask[:, None].repeat(1, X.shape[1], 1) 

	X_var = X[mask].reshape(X.shape[0], X.shape[1], -1)

	if left == True:
		X = X[:, :, -X_var.shape[-1]:]
	else:
		X = X[:, :, :X_var.shape[-1]]

	y_before = func(model, X, args=args, **additional_func_kwargs, **kwargs)
	y_after = func(model, X_var, args=args, **additional_func_kwargs, **kwargs)
	return y_before, y_after


def insertion_effect(model, X, insertions, left=False, args=None, func=predict,
	additional_func_kwargs=None, **kwargs):
	"""Apply a function before and after inserting characters into a sequence.

	This function will calculate the effect that insertions have on the
	output from a model. Any number of insertions can be specified for each
	sequence and the provided `func` is applied before and after the insertions
	are taken into account. By default, this `func` is the prediction function
	and so the results are the difference in predictions before and after the
	insertions are made, but if `func` is something else, such as 
	`deep_lift_shap`, the results will be the attributions before and after the
	insertions are made. At least one insertion should be provided per sequence
	for the results to be different.

	The insertions provided must be individual characters, i.e., that each row
	in the tensor corresponds to a single insertion in a single example, but one
	can encode longer insertions (e.g., entire motifs or just multiple 
	characters) by passing in multiple rows with adjacent positions.

	Simply removing all of the specified characters will lead to a set of
	sequences of differing lengths if there are a different number of insertions
	in each sequence. In order to make these sequences all the same length, we
	need to trim from each sequence a number of positions from each sequence
	equal to the number of characters that are being added in. Because there
	are two ways we can trim positions -- either starting from the left end of 
	the sequence or from the right end of it -- you can use the `left` parameter
	to specify that you should trim positions on the left (`left=True`) or from
	the right (`left=False`, the default) of the sequence. 


	Parameters
	----------
	model: torch.nn.Module
		A PyTorch model to use for making predictions. These models can take in
		any number of inputs and make any number of outputs. The additional
		inputs must be specified in the `args` parameter.

	X: torch.tensor, shape=(-1, len(alphabet), length)
		A one-hot encoded set of sequences to have insertions included in.

	insertions: torch.tensor, shape=(-1, 3)
		A set of insertions indicating characters that should be added to
		each sequence. Each row should be a single insertion with the first
		column corresponding to the example and the second column corresponding
		to the position within that example. Multiple insertions can occur in 
		each example. 

	left: bool, optional
		If False, use the first `n` positions to run through the model before
		making the deletion where `n` is the expected tensor length. If True,
		use the last `n` positions. Basically, whether we trim positions from
		the left or the right when getting sequences of the same length.
		Default is False.

	args: tuple or None, optional
		An optional set of additional arguments to pass into the model. If
		provided, each element in the tuple or list is one input to the model
		and the element must be formatted to be the same batch size as `X`. If
		None, no additional arguments are passed into the forward function. This
		argument is provided here because the args must be copied for each
		shuffle that occurs. Default is None.

	func: function, optional
		A function to apply before and after incorporating the substitution. 
		Default is `predict`.

	additional_func_kwargs: dict, optional
		Additional named arguments to pass into the function when it is called.
		This is provided as an alternate path to route arguments into the 
		function in case they overlap, name-wise, with those in this function,
		or if you want to be absolutely sure that the arguments are making
		their way into the function. Default is {}.

	kwargs: optional
		Additional named arguments that will get passed into the function when
		it is called. Default is no arguments are passed in.


	Returns
	-------
	y_before: torch.Tensor
		The output from `func` before variants are included.

	y_after: torch.Tensor
		The output from `func` after the variants are included.
	"""

	insertions = _cast_as_tensor(insertions)

	additional_func_kwargs = additional_func_kwargs or {}
	X_var = []

	for i in range(X.shape[0]):
		insertions_ = insertions[insertions[:, 0] == i]
		insertions_ = insertions_[torch.argsort(insertions_[:, 1], 
			descending=True)]

		x = X[i:i+1]
		for _, j, char in insertions_:
			v = torch.zeros(1, X.shape[1], 1)
			v[:, char] = 1
			x = insert(x, v, start=j)

		if left == True:
			x = x[:, :, -X.shape[-1]:]
		else:
			x = x[:, :, :X.shape[-1]]
		
		X_var.append(x)

	X_var = torch.cat(X_var)
	y_before = func(model, X, args=args, **additional_func_kwargs, **kwargs)
	y_after = func(model, X_var, args=args, **additional_func_kwargs, **kwargs)
	return y_before, y_after



# match.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>
# Contribution by: Bo Vagner Hansen <bvh@bmb.sdu.dk>

"""
Provides functions for the calculation of GC-content genome-wide and the
sampling of GC-matched negatives.
"""

import numpy
import pandas
import pyfaidx
import pyBigWig

from tqdm import tqdm
from scipy.stats import ks_2samp
from joblib import Parallel, delayed

def _get_chrom_sizes_dict(fasta, chroms):
	"""Returns a dictionary with the elements of `chroms` as keys and the
	size of the chromosomes as values, extracted from the file `fasta`."""
	with pyfaidx.Fasta(fasta) as genome_stream:
		return {chrom:len(genome_stream[chrom]) for chrom in chroms}

def _chrom_coords_generator(chrom, chrom_size, width):
	"""Tiles a given `chrom` from 0 to `chrom_size`, returning region
	coords of size `width`. Produces chrom_size // width regions,
	ignoring the remainder."""
	start = 0
	for end in range(width, chrom_size+1, width):
		yield chrom, start, end
		start = end

def _resize_coords_generator(coords, width):
	"""Resizes the given `coords` to have size `width` and the same midpoint."""
	left_flank = width // 2
	right_flank = (width+1) // 2
	for chrom, start, end in coords:
		mid = start + (end-start) // 2
		yield chrom, mid-left_flank, mid+right_flank

def _loci_coords_generator(loci_df, width = None):
	"""Takes in a pandas dataframe `loci_df` and returns a generator of loci coordinates.
	The dataframe must contain columns named chrom, start and end - other columns are not
	used. If `width` is set to None, will use the start and end coordinates as given.
	if `width` is an integer, will return start and end coordinates that are `width` apart,
	and centered on the midpoint of the given start and end coordinates."""
	g = ((locus.chrom, locus.start, locus.end) for locus in loci_df.itertuples(index=False))
	return g if width is None else _resize_coords_generator(g, width)

def _valid_locus(chrom, start, end, chrom_sizes):
	"""Returns a bool telling whehter the specificed locus is valid, i.e the `chrom`
	is found in the `chrom_sizes` dictionary and `start` and `end` are within bounds."""
	return (chrom in chrom_sizes) and (start >= 0) and (end <= chrom_sizes[chrom])

def _valid_generator(coords, chrom_sizes):
	"""Returns a generator filtering the provided `coords` for invalid loci using
	the `chrom_sizes` dictionary."""
	return (locus for locus in coords if _valid_locus(*locus, chrom_sizes))

def _sequence_generator_slice(coords, genome):
	"""Generates the sequences given by `coords` from the sliceable `genome`,
	e.g. a str, list or tuple."""
	return (genome[start:end] for _,start,end in coords)

def _sequence_generator_stream(coords, genome_stream):
	"""Streams the sequences given by `coords` from the pyfaidx `genome_stream`.
	Returns a generator, so that only one sequence is kept in memory at a time
	when passing the generator on to an iterator that consumes the sequence."""
	return (genome_stream[chrom][start:end].seq.upper() for chrom,start,end in coords)

def _sequence_generator_buffered(coords, genome_stream):
	"""Returns a generator of sequences given by `coords` from the pyfaidx `genome_stream`.
	The entire chromosomal sequence is loaded into memory. Should mainly be used when
	there are many coords on each chromosome. If not all coords are on the same chromosome,
	the coords should definitely be sorted, as otherwise the function will be very inefficient."""
	buffer_chrom = ''
	for chrom, start, end in coords:
		if buffer_chrom != chrom:
			buffer_chrom = chrom
			buffer_seq = genome_stream[chrom][:].seq
		yield buffer_seq[start:end].upper()

def _sequence_generator(coords, genome, buffer = False):
	"""Wrapper function returning the sequence generator corresponding to the value of
	`genome` and `buffer`. Takes in a `genome`, which can be either a pyfaidx.Fasta object
	or a sliceable object like a str, list or tuple. `buffer` is only used for pyfaidx.Fasta
	objects. `buffer` determines whether the sequences should be streamed (False) or extracted
	as slices from a buffered chromosomal sequence (True)."""
	if isinstance(genome, pyfaidx.Fasta):
		if buffer:
			return _sequence_generator_buffered(coords, genome)
		else:
			return _sequence_generator_stream(coords, genome)
	else:
		return _sequence_generator_slice(coords, genome)

def _perc_generator(sequences, chars):
	"""Calculates the percentage (as a decimal) of `chars` in the provided `sequences`.
	Returns a generator of these percentages."""
	return (sum(seq.count(c) for c in chars) / len(seq) for seq in sequences)

def _extract_counts(chrom, start, end, bw_stream):
	"""Extract the signal from the pyBigWig `bw_stream` in the provided locus and returns
	the sum of the base pair values. If the signal cannot be extracted, returns nan."""
	try:
		value = bw_stream.stats(chrom, start, end, type="sum", exact=True)[0]
		return value if value is not None else 0
	except:
		return float('nan')

def _count_generator(coords, bw_stream, buffer = False):
	"""Wrapper function returning the count generator corresponding to the value of `buffer`.
	If no signal can be extracted from a locus, nan is returned for that locus.
	Buffer determines whether the signals should be streamed (False) or extracted as slices
	from a buffered chromosomal signal (True)."""
	if buffer:
		return _count_generator_buffered(coords, bw_stream)
	else:
		return _count_generator_stream(coords, bw_stream)

def _count_generator_stream(coords, bw_stream):
	"""Takes a list of locus `coords` and a pyBigWig `bw_stream` and returns a generator
	producing the sum of counts for each locus. The signals are streamed from the locus,
	such that only one locus signal is kept in memory at a time. As reading from a bigwig
	is a bit slow, it can be quite time consuming to use this function across an entire
	chromosome, hence the buffered version should be preferred in that case."""
	return (_extract_counts(*locus, bw_stream) for locus in coords)

def _count_generator_buffered(coords, bw_stream):
	"""Takes a list of locus `coords` and a pyBigWig `bw_stream` and returns a generator
	producing the sum of counts for each locus. The signal for an entire chromosome
	is kept in memory at a time, to speed up the extraction of multiple signals across
	a chromosome. This function should mainly be used if calculating the counts for
	densely spaced regions across the chromosomes. In addition, `coords` should be sorted
	by chromosome, otherwise the function will be very inefficient."""
	buffer_chrom = ''
	for chrom, start, end in coords:
		if buffer_chrom != chrom:
			buffer_chrom = chrom
			try:
				buffer_signal = bw_stream.values(chrom, 0, -1, numpy=True)
			except:
				buffer_signal = None
		if buffer_signal is not None:
			yield numpy.nansum(buffer_signal[start:end]).item()
		else:
			yield float('nan')

def _char_perc_from_coords(fasta, coords, chars, num_regions=-1, buffer=False, verbose=False):
	"""This method will take in a `fasta` file and return the percentage of `chars`
	in the sequences extracted from the fasta file and defined by the list of `coords`.
	This is usually used to calculate GC percentage but can also be used to calculate
	the percentage of N's.

	Parameters
	----------
	fasta: str
		The path to a fasta file, usually a reference genome.

	coords: list, tuple or generator of such
		iterable of tuples formatted like (chr, start, end),
		where `chr` is a string and `start` and `end` are integers.

	chars: iterable, such as set, list, tuple or str
		The characters to look for in the sequences bounded by `coords`.
		The function returns the percentage of these characters found in
		each sequence.

	num_regions: int, optional
		The number of regions given by coords, i.e. the length of coords.
		Used for efficient construction of the output numpy array, in case
		coords is given as a generator. If set to -1, the number or regions
		will automatically be inferred. Default is -1.

	buffer: bool, optional
		Whether to load the entire chromosomal sequence into memory and
		and extract the sequences as slices of the entire chromosome.
		Should only be set to true if there are many regions
		on the same chromosome, such as when calculating GC content for
		an entire chromosome. It is not recommend to use buffering if
		calculating char pecentages for only peak regions, but if done
		anyway for some reason, the peaks should be sorted by chromosome.
		If set to false, will instead only load the sequence for a single
		region at a time into memory. Default is False.

	verbose: bool, optional
		Whether to display a progress bar.

	Returns
	-------
	perc: numpy.ndarray, shape=(len(list(coords)), )
	"""
	desc = "Getting %s percentages" % ''.join(chars)
    
	with pyfaidx.Fasta(fasta) as genome_stream:
		generator = _sequence_generator(coords, genome_stream, buffer=buffer)
		generator = _perc_generator(generator, chars)
		generator = tqdm(generator, disable=not verbose, desc=desc)
		perc = numpy.fromiter(generator, dtype=float, count=num_regions)

	return perc

def _counts_from_coords(bigwig, coords, num_regions=-1, buffer=False, verbose=False):
	"""An internal function returning the percentage of `chars` for each sequence
	with `coords` extracted from the `fasta` file.

	This method will take in a `fasta` file and return the percentage of `chars`
	in the sequences extracted from the fasta file and defined by the list of `coords`.
	This is usually used to calculate GC percentage but can also be used to calculate
	the percentage of N's.

	Parameters
	----------
	fasta: str
		The path to a bigwig file to extract counts from.

	coords: list, tuple or generator of such
		iterable of tuples formatted like (chr, start, end),
		where `chr` is a string and `start` and `end` are integers.

	num_regions: int, optional
		The number of regions given by coords, i.e. the length of coords.
		Used for efficient construction of the output numpy array, in case
		coords is given as a generator. If set to -1, the number or regions
		will automatically be inferred. Default is -1.

	buffer: bool, optional
		Whether to load the entire chromosomal signal into memory and
		and extract the locus signals as slices of the entire chromosome.
		Should only be set to true if there are many regions on the same 
		chromosome. If calculating counts for regions on many chromosomes,
		the regions should be sorted by chromosome.
		If set to false, will instead only load the signal for a single
  		region at a time into memory. Default is False.

	verbose: bool, optional
		Whether to display a progress bar.

	Returns
	-------
	count: numpy.ndarray, shape=(len(list(coords)), )
	"""
	desc = "Getting counts"

	with pyBigWig.open(bigwig, "r") as bw_stream:
		generator = _count_generator(coords, bw_stream, buffer=buffer)
		generator = tqdm(generator, disable=not verbose, desc=desc)
		count = numpy.fromiter(generator, dtype=float, count=num_regions)

	return count

def _calculate_char_perc(sequence, width, chars):
	"""An internal function returning the percentage of `chars` in `sequence`.

	This method will take in a string `sequence` and return the percentage
	of each non-overlapping block of length `width` across the sequence
	containing `chars`. This is usually used to calculate GC percentage but
	can also be used to calculate the percentage of N's in a sequence.
	

	Parameters
	----------
	sequence: str
		A string made up of the alphabet 'A', 'C', 'G', 'T', 'N'.

	width: int
		The total width of the window to calculate the GC content for, e.g.,
		`1000` if you want to calculate this for 1000 bp blocks.

	chars: iterable, such as set, list, tuple or str
		The characters to look for in the sequence. The returned percentage is
		the percentage of the original sequence that is one of these characters.

	Returns
	-------
	perc: numpy.ndarray, shape=(len(sequence) // width,)
	"""

	seq_len = len(sequence)
	num_regions = seq_len // width
	coords = _chrom_coords_generator(chrom='', chrom_size=seq_len, width=width)
	seqs = _sequence_generator(coords, sequence)
	perc = _perc_generator(seqs, chars)
	perc = numpy.fromiter(perc, dtype=float, count=num_regions)
	return perc

def _extract_and_filter_chrom(fasta, chrom, in_window, out_window, 
	max_n_perc=0.1, gc_bin_width=0.02, bigwig=None, signal_threshold=None):
	"""Calculate GC content for, and filter, one chromosome.

	This function will take in the name of a FASTA file, a chromosome, and a
	percentage of Ns that cannot be exceeded, and return a set of passing loci
	and their exact GC percentage. Optionally, it will also take in a bigwig
	and filter the loci based on having a signal threshold smaller than some
	value, where the signal is summed across the width.

	
	Parameters
	----------
	fasta: str
		The filepath to the FASTA file to extract sequences from.

	chrom: str
		The chromosome to extract from the FASTA file. Must be in the file.

	in_window: int
		The window to calculate the GC content over, corresponding to the input
		window of the downstream model that will be trained.

	out_window: int
		The window to calculate signal for and apply the signal threshold to,
		corresponding to the output window of the downstream model that will
		be trained.

	max_n_perc: float, range=(0, 1.0), optional
		The maximum percentage of N characters in each window to be considered.
		All windows with a higher percentage are discarded. Default is 0.1.

	gc_bin_width: float, range=(0, 1.0), optional
		The bin size to discretize GC content. Default is 0.02.

	bigwig: str or None, optional
		If filtering regions based on signal strength, calculate the signal
		from this bigwig. If None, do not filter based on signal strength.
		Default is None.

	signal_threshold: float or None, optional
		The maximum possible signal, summed across the entire window, that
		each window can have without being filtered. If the window has a summed
		signal higher than this value, the window is discarded. Default is None.

	Returns
	-------
	gc_percs: dict
		A dictionary where the keys are observed GC bins and the values are
		lists of loci that are in that GC bin. Each returned locus is the index
		not the true value and so corresponds to the real position integer
		divided by in_window.
	"""

	with pyfaidx.Fasta(fasta) as f:
		sequence = f[chrom][:].seq.upper()

	gc_perc = _calculate_char_perc(sequence, in_window, 'GC')
	n_perc = _calculate_char_perc(sequence, in_window, 'N')
	del sequence

	idxs = n_perc <= max_n_perc
    
	if bigwig is not None:
		assert(in_window >= out_window), "out_window cannot be larger than in_window."
		left_flank = (in_window - out_window) // 2
		right_flank = (in_window - out_window + 1) // 2
        
		with pyBigWig.open(bigwig, "r") as bw:
			try:
				values = bw.values(chrom, 0, -1, numpy=True)
			except RuntimeError:
				return {}
        
		values = values[:values.shape[0] // in_window * in_window]
		values = values.reshape(-1, in_window)
		values = values[:, left_flank:-right_flank]
		values = numpy.nansum(values, axis=-1)

		idxs = idxs & (values <= signal_threshold)

	gc_perc = ((gc_perc + gc_bin_width / 2.) // gc_bin_width).astype(int)
	unique_gc = numpy.unique(gc_perc[idxs]).tolist()
	gc_perc = {gc:numpy.nonzero(idxs & (gc_perc == gc))[0].tolist() for gc in unique_gc}

	return gc_perc


def extract_matching_loci(loci, fasta, in_window=2114, out_window=1000, 
	max_n_perc=0.1, gc_bin_width=0.02, bigwig=None, signal_beta=0.5, 
	chroms=None, random_state=None, n_jobs=-1, verbose=False):
	"""Extract matching loci given a fasta file.

	This function takes in a set of loci (a bed file or a pandas dataframe in 
	bed format) and returns a GC-matched set of negatives. This will also
	perform basic filtering to ignore regions of the genome that are too high
	in Ns. Optionally, it can take in a bigwig and a signal threshold and only
	select regions that have fewer than a threshold of counts in each region.

	Importantly, it will apply `max_n_perc` to both the loci that are passed in
	and also potential regions that can be selected. This means that if a locus
	passed in has higher than `max_n_perc` number of unspecified positions,
	it will be filtered out, and a smaller number of positions will be selected.
	This is done because the GC content of a region with many Ns in it is not
	trustworthy.


	Parameters
	----------
	loci: str or pandas dataframe
		A filepath to a bed file, or a pandas dataframe in bed format.

	fasta: str
		The filepath to the FASTA file to extract sequences from.

	in_window: int
		The window to calculate the GC content over, corresponding to the input
		window of the downstream model that will be trained.

	out_window: int
		The window to calculate signal for and apply the signal threshold to,
		corresponding to the output window of the downstream model that will
		be trained.

	max_n_perc: float, range=(0, 1.0), optional
		The maximum percentage of N characters in each window to be considered.
		All windows with a higher percentage are discarded. Default is 0.1.

	gc_bin_width: float, range=(0, 1.0), optional
		The bin size to discretize GC content. Default is 0.02.

	bigwig: str or None, optional
		If filtering regions based on signal strength, calculate the signal
		from this bigwig. If None, do not filter based on signal strength.
		Default is None.

	signal_beta: float or None, optional
		A multiplier of the robust minimum signal calculated from `loci` that
		each background region must have fewer reads then. Only relevant if a
		bigwig is passed in. Default is 0.5.

	chroms: list, tuple, or None, optional
		A set of chromosomes to use when choosing matching loci. If None, only
		use chromosomes that the loci themselves are drawn from. Default is
		None.

	random_state: numpy.random.RandomState, int or None, optional
		A random state to use for sampling loci. If a RandomState object or
		an integer, this will produce deterministic sampling. If None, sampling
		will be different each time. Default is None.

	n_jobs: integer, optional
		Number of parallel processes to use for extracting background gc content.
		-1 means use all available CPUs. Default is -1. 

	verbose: bool, optional
		Whether to print display bars and diagnostics to ensure that the
		sampling is reasonable. When set to True, there may be a large amount
		of output. Default is False. 


	Returns
	-------
	matched_loci: pandas.DataFrame
		A bed-formatted set of matched loci sorted first by chromosome and
		then by position on the chromosome. Note that these are not sorted
		such that the i-th position in this file is a GC match for the i-th
		position in the original locus file.
	"""

	if not isinstance(random_state, numpy.random.RandomState):
		random_state = numpy.random.RandomState(random_state)

	if isinstance(loci, str):
		loci = pandas.read_csv(loci, sep='\t', usecols=[0, 1, 2], header=None, 
			index_col=False, names=['chrom', 'start', 'end'])

	loci_chroms = numpy.unique(loci['chrom'])
	if chroms is None:
		chroms = loci_chroms

	chrom_sizes = _get_chrom_sizes_dict(fasta, loci_chroms)

	if verbose: print("Processing given loci.")
	coords = _loci_coords_generator(loci, max(in_window, out_window))
	coords = list(_valid_generator(coords, chrom_sizes))
	num_regions = len(coords)

	threshold = None
	if bigwig is not None:
		coords = list(_resize_coords_generator(coords, out_window))
		loci_count = _counts_from_coords(bigwig, coords, num_regions, buffer=False, verbose=verbose)
		robust_min = numpy.nanquantile(loci_count, 0.01).item()
		threshold = robust_min * signal_beta
    
	coords = list(_resize_coords_generator(coords, in_window))
	loci_n = _char_perc_from_coords(fasta, coords, 'N', num_regions, buffer=False, verbose=verbose)
	loci_gc = _char_perc_from_coords(fasta, coords, 'GC', num_regions, buffer=False, verbose=verbose)
	loci_gc = loci_gc[loci_n < max_n_perc]

	loci_gc = ((loci_gc + gc_bin_width / 2.) // gc_bin_width).astype(int)
	loci_bin_count = numpy.zeros(int(1./gc_bin_width)+1, dtype=int)
	for gc_bin in loci_gc:
		loci_bin_count[gc_bin] += 1

	# Extract mask of already-selected loci
	mask = {chrom: [] for chrom in chroms}
	for locus in loci.itertuples(index=False):
		if locus.chrom not in mask:
			continue

		start = locus.start // in_window
		end = locus.end // in_window + 1

		mask[locus.chrom].extend(range(start, end))

	for chrom, values in mask.items():
		mask[chrom] = set(values)

	# Get GC content of background regions
	desc = 'Getting background GC'
	f = delayed(_extract_and_filter_chrom)
	chrom_percs = Parallel(n_jobs=n_jobs)(f(
		fasta=fasta, 
		chrom=chrom, 
		in_window=in_window,
		out_window=out_window, 
		max_n_perc=max_n_perc,
		gc_bin_width=gc_bin_width,
		bigwig=bigwig, 
		signal_threshold=threshold) 
		for chrom in tqdm(chroms, disable=not verbose, desc=desc))
   
	# Merge them into a single dictionary, keeping track of chroms
	bg_bin_count = numpy.zeros(int(1./gc_bin_width) + 1, dtype=int)
	gc_percs = {perc: [] for perc in range(len(bg_bin_count))}
	
	for chrom, percs in zip(chroms, chrom_percs):
		for key, values in percs.items():
			for value in values:
				if value not in mask[chrom]:
					gc_percs[key].append((chrom, value))
					bg_bin_count[key] += 1

	for key, value in gc_percs.items():
		random_state.shuffle(value)

	orig_bg_bin_count = bg_bin_count.copy()
	orig_loci_bin_count = loci_bin_count.copy()

	# Match the sizes
	matched_loci_bin_count = numpy.minimum(bg_bin_count, loci_bin_count)
	bg_bin_count -= matched_loci_bin_count
	loci_bin_count -= matched_loci_bin_count

	n = len(loci_bin_count)
	for i in range(n-1, -1, -1):
		if loci_bin_count[i] == 0:
			continue

		for offset in range(n):
			idx = i + offset
			if idx < n:
				count = min(bg_bin_count[idx], loci_bin_count[i])
				bg_bin_count[idx] -= count
				loci_bin_count[i] -= count
				matched_loci_bin_count[idx] += count

			if loci_bin_count[i] == 0:
				break

			idx = i - offset
			if idx > 0:
				count = min(bg_bin_count[idx], loci_bin_count[i])
				bg_bin_count[idx] -= count
				loci_bin_count[i] -= count
				matched_loci_bin_count[idx] += count

			if loci_bin_count[i] == 0:
				break

	if verbose:
		numpy.set_printoptions(suppress=True)
		print("GC Bin\tBackground Count\tPeak Count\tChosen Count")
		for i in range(n):
			print("{:2.2f}: {:8d}\t{:8d}\t{:8d}".format(
				numpy.arange(0, 1.01, gc_bin_width)[i], 
				orig_bg_bin_count[i], orig_loci_bin_count[i], 
				matched_loci_bin_count[i]))

	# Extract the loci
	matched_loci = {'chrom': [], 'start': [], 'end': []}
	for i in range(n):
		for j in range(matched_loci_bin_count[i]):
			chrom, start = gc_percs[i][j]

			matched_loci['chrom'].append(chrom)
			matched_loci['start'].append(start*in_window)
			matched_loci['end'].append((start+1)*in_window)

	matched_loci = pandas.DataFrame(matched_loci)

	if verbose:     
		matched_gc = []
		for i,j in enumerate(matched_loci_bin_count):
			matched_gc.extend([i]*j)
            
		stats = ks_2samp(loci_gc, matched_gc)
		print("GC-bin KS test stat:{:3.3}, p-value {:3.3}".format(
			stats.statistic, stats.pvalue))

		if bigwig is not None:
			print("Processing matched loci.")
			coords = _loci_coords_generator(matched_loci, out_window)
			num_regions = len(matched_loci)
			matched_count_max = _counts_from_coords(bigwig, coords, num_regions, buffer=False, verbose=verbose).max()
			print("Peak Robust Signal Minimum: {}".format(robust_min))
			print("Matched Signal Maximum: {}".format(matched_count_max))
  
	matched_loci = matched_loci.sort_values(["chrom", "start"])
	return matched_loci



# test_io.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch
import pytest
import collections

from tangermeme.utils import characters
from tangermeme.utils import one_hot_encode
from tangermeme.utils import random_one_hot

from tangermeme.ersatz import insert
from tangermeme.ersatz import substitute
from tangermeme.ersatz import multisubstitute
from tangermeme.ersatz import delete
from tangermeme.ersatz import randomize
from tangermeme.ersatz import shuffle
from tangermeme.ersatz import dinucleotide_shuffle

import pandas
import pyfaidx
import pyBigWig

from numpy.testing import assert_raises
from numpy.testing import assert_array_almost_equal


@pytest.fixture
def X():
	seq = 'CACATCATCTCATCATCTGCTGACTACTGACGTAGTCTGACTGACTGACTGACTACTGACTGACTGAC'
	return one_hot_encode(seq).unsqueeze(0)


###


def test_insert_str(X):
	motif = 'CATCAG'
	X_insert = insert(X, motif)

	assert X_insert.shape[:2] == X.shape[:2]
	assert X_insert.shape[-1] != X.shape[-1]
	assert X_insert.sum() != X.sum()

	assert_raises(AssertionError, assert_array_almost_equal, 
		X_insert.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_insert, X)

	new_seq = ('CACATCATCTCATCATCTGCTGACTACTGACGTACAT' + 
		'CAGGTCTGACTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)

	assert X_insert.shape == new_seq_ohe.shape
	assert X_insert.sum() == new_seq_ohe.sum()
	assert_array_almost_equal(X_insert.sum(dim=-1), new_seq_ohe.sum(dim=-1))
	assert_array_almost_equal(X_insert, new_seq_ohe)


def test_insert_ohe(X):
	motif = one_hot_encode('CATCAG').unsqueeze(0)
	X_insert = insert(X, motif)

	assert X_insert.shape[:2] == X.shape[:2]
	assert X_insert.shape[-1] != X.shape[-1]
	assert X_insert.sum() != X.sum()

	assert_raises(AssertionError, assert_array_almost_equal, 
		X_insert.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_insert, X)

	new_seq = ('CACATCATCTCATCATCTGCTGACTACTGACGTACAT' + 
		'CAGGTCTGACTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)

	assert X_insert.shape == new_seq_ohe.shape
	assert X_insert.sum() == new_seq_ohe.sum()
	assert_array_almost_equal(X_insert.sum(dim=-1), new_seq_ohe.sum(dim=-1))
	assert_array_almost_equal(X_insert, new_seq_ohe)


def test_insert_start(X):
	motif = 'CATCAG'
	X_insert = insert(X, motif, start=0)

	assert X_insert.shape[:2] == X.shape[:2]
	assert X_insert.shape[-1] != X.shape[-1]
	assert X_insert.sum() != X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_insert.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_insert, X)

	new_seq = ('CATCAGCACATCATCTCATCATCTGCTGACTAC' + 
		'TGACGTAGTCTGACTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)
	assert_array_almost_equal(X_insert, new_seq_ohe)


	motif = 'CATCAGCCC'
	X_insert = insert(X, motif, start=10)

	assert X_insert.shape[:2] == X.shape[:2]
	assert X_insert.shape[-1] == 77
	assert X_insert.sum() != X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_insert.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_insert, X)

	new_seq = ('CACATCATCTCATCAGCCCCATCATCTGCTGACTACTGA' + 
		'CGTAGTCTGACTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)
	assert_array_almost_equal(X_insert, new_seq_ohe)


def test_insert_long(X):
	motif = 'C'*1000
	X_insert = insert(X, motif)
	assert X_insert.shape == (1, 4, 1068)
	assert X_insert[0, 1].sum() > 1000


def test_insert_raises_alphabet(X):
	motif = one_hot_encode('CACCAG', alphabet=['A', 'C', 'G']).unsqueeze(0)
	assert_raises(ValueError, insert, X, motif)
	assert_raises(ValueError, insert, X, motif, None, ['A', 'C', 'G'])


def test_insert_raise_ends(X):
	assert_raises(ValueError, insert, X, 'CAGCAT', -2)
	assert_raises(ValueError, insert, X, 'CAGCAT', 1000)
	assert_raises(TypeError, insert, X, 'CAGCAT', 6.5)
	assert_raises(TypeError, insert, X, 'CAGCAT', 5, 3)
	assert_raises(TypeError, insert, X, 'CAGCAT', -5, -1)
	assert_raises(TypeError, insert, X, 'CAGCAT', 5, 1000)


def test_insert_raises_ohe(X):
	motif = one_hot_encode('CATCAG').unsqueeze(0)
	motif[0, 0, 0] = 1
	assert_raises(ValueError, insert, X, motif)

	motif = one_hot_encode('CATCAG').unsqueeze(0)
	motif[0, 1, 0] = 2
	assert_raises(ValueError, insert, X, motif)
	assert_raises(ValueError, insert, X, torch.randn(1, 4, 8))


###


def test_substitute_str(X):
	motif = 'CATCAG'
	X_substitute = substitute(X, motif)

	assert X_substitute.shape == X.shape
	assert X_substitute.sum() == X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_substitute.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_substitute, X)

	new_seq = ('CACATCATCTCATCATCTGCTGACTACTGACCAT' + 
		'CAGTGACTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)

	assert X_substitute.shape == new_seq_ohe.shape
	assert X_substitute.sum() == new_seq_ohe.sum()
	assert_array_almost_equal(X_substitute.sum(dim=-1), new_seq_ohe.sum(dim=-1))
	assert_array_almost_equal(X_substitute, new_seq_ohe)


def test_substitute_ohe(X):
	motif = one_hot_encode('CATCAG').unsqueeze(0)
	X_substitute = substitute(X, motif)

	assert X_substitute.shape == X.shape
	assert X_substitute.sum() == X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_substitute.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_substitute, X)

	new_seq = ('CACATCATCTCATCATCTGCTGACTACTGACCAT' + 
		'CAGTGACTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)

	assert X_substitute.shape == new_seq_ohe.shape
	assert X_substitute.sum() == new_seq_ohe.sum()
	assert_array_almost_equal(X_substitute.sum(dim=-1), new_seq_ohe.sum(dim=-1))
	assert_array_almost_equal(X_substitute, new_seq_ohe)


def test_substitute_str_multi_seqs_one_motif():
	X = random_one_hot((4, 4, 8), random_state=0)
	X_substitute = substitute(X, 'ACGT')

	assert_raises(AssertionError, assert_array_almost_equal, X, X_substitute)
	assert_array_almost_equal(X_substitute, [
		[[1, 0, 1, 0, 0, 0, 0, 0],
         [0, 0, 0, 1, 0, 0, 0, 0],
         [0, 0, 0, 0, 1, 0, 0, 0],
         [0, 1, 0, 0, 0, 1, 1, 1]],

        [[0, 0, 1, 0, 0, 0, 0, 1],
         [1, 0, 0, 1, 0, 0, 0, 0],
         [0, 0, 0, 0, 1, 0, 1, 0],
         [0, 1, 0, 0, 0, 1, 0, 0]],

        [[1, 1, 1, 0, 0, 0, 0, 0],
         [0, 0, 0, 1, 0, 0, 0, 0],
         [0, 0, 0, 0, 1, 0, 0, 1],
         [0, 0, 0, 0, 0, 1, 1, 0]],

        [[1, 0, 1, 0, 0, 0, 0, 1],
         [0, 1, 0, 1, 0, 0, 1, 0],
         [0, 0, 0, 0, 1, 0, 0, 0],
         [0, 0, 0, 0, 0, 1, 0, 0]]])


def test_substitute_str_multi_seqs_multi_motifs():
	X = random_one_hot((4, 4, 8), random_state=0)
	motif = random_one_hot((4, 4, 4), random_state=1)
	X_substitute = substitute(X, motif)

	assert_raises(AssertionError, assert_array_almost_equal, X, X_substitute)
	assert_array_almost_equal(X_substitute, [
		[[1, 0, 0, 0, 1, 1, 0, 0],
         [0, 0, 1, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 0],
         [0, 1, 0, 1, 0, 0, 1, 1]],

        [[0, 0, 0, 0, 0, 0, 0, 1],
         [1, 0, 0, 1, 0, 1, 0, 0],
         [0, 0, 0, 0, 0, 0, 1, 0],
         [0, 1, 1, 0, 1, 0, 0, 0]],

        [[1, 1, 0, 1, 1, 0, 0, 0],
         [0, 0, 0, 0, 0, 1, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 1],
         [0, 0, 1, 0, 0, 0, 1, 0]],

        [[1, 0, 1, 0, 0, 1, 0, 1],
         [0, 1, 0, 0, 1, 0, 1, 0],
         [0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 1, 0, 0, 0, 0]]])


def test_substitute_start_str(X):
	motif = 'CATCAG'
	X_substitute = substitute(X, motif, start=0)

	assert X_substitute.shape == X.shape
	assert X_substitute.sum() == X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_substitute.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_substitute, X)

	new_seq = ('CATCAGATCTCATCATCTGCTGACTACT' + 
		'GACGTAGTCTGACTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)
	assert_array_almost_equal(X_substitute, new_seq_ohe)


	motif = 'CATCAGCCC'
	X_substitute = substitute(X, motif, start=10)

	assert X_substitute.shape == X.shape
	assert X_substitute.sum() == X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_substitute.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_substitute, X)

	new_seq = ('CACATCATCTCATCAGCCCCTGACTACTGACGTA' +
		'GTCTGACTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)
	assert_array_almost_equal(X_substitute, new_seq_ohe)


def test_substitute_start_ohe(X):
	motif = one_hot_encode('CATCAG').unsqueeze(0)
	X_substitute = substitute(X, motif, start=0)

	assert X_substitute.shape == X.shape
	assert X_substitute.sum() == X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_substitute.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_substitute, X)

	new_seq = ('CATCAGATCTCATCATCTGCTGACTACT' + 
		'GACGTAGTCTGACTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)
	assert_array_almost_equal(X_substitute, new_seq_ohe)


	motif = one_hot_encode('CATCAGCCC').unsqueeze(0)
	X_substitute = substitute(X, motif, start=10)

	assert X_substitute.shape == X.shape
	assert X_substitute.sum() == X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_substitute.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_substitute, X)

	new_seq = ('CACATCATCTCATCAGCCCCTGACTACTGACGTA' +
		'GTCTGACTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)
	assert_array_almost_equal(X_substitute, new_seq_ohe)


def test_substitute_raises_alphabet(X):
	motif = one_hot_encode('CACCAG', alphabet=['A', 'C', 'G'])
	assert_raises(ValueError, substitute, X, motif)
	assert_raises(ValueError, substitute, X, motif, ['A', 'C', 'G'])


def test_substitute_raises_length(X):
	assert_raises(ValueError, substitute, X, 'C'*1000)
	assert_raises(ValueError, substitute, X, one_hot_encode('C'*1000))


def test_substitute_raise_ends(X):
	assert_raises(ValueError, substitute, X, 'CAGCAT', -2)
	assert_raises(ValueError, substitute, X, 'CAGCAT', 1000)
	assert_raises(TypeError, substitute, X, 'CAGCAT', 6.5)
	assert_raises(TypeError, substitute, X, 'CAGCAT', 5, 3)
	assert_raises(TypeError, substitute, X, 'CAGCAT', -5, -1)
	assert_raises(TypeError, substitute, X, 'CAGCAT', 5, 1000)


def test_substitute_raises_ohe(X):
	motif = one_hot_encode('CATCAG').unsqueeze(0)
	motif[0, 0, 0] = 1
	assert_raises(ValueError, substitute, X, motif)

	motif = one_hot_encode('CATCAG').unsqueeze(0)
	motif[0, 1, 0] = 2
	assert_raises(ValueError, substitute, X, motif)
	assert_raises(ValueError, substitute, X, torch.randn(1, 4, 8))


###


def test_delete(X):
	X_delete = delete(X, start=0, end=5)

	assert X_delete.shape != X.shape
	assert X_delete.shape[-1] == X.shape[-1] - 5
	assert X_delete.sum() == X[:, :, 5:].sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_delete.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_delete, X)

	X_delete = delete(X, start=10, end=17)

	assert X_delete.shape != X.shape
	assert X_delete.shape[-1] == X.shape[-1] - 7
	assert_array_almost_equal(X_delete[:, :, :10], X[:, :, :10])
	assert_array_almost_equal(X_delete[:, :, 10:], X[:, :, 17:])

	assert_raises(AssertionError, assert_array_almost_equal, 
		X_delete.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_delete, X)



def test_substitute_raise_ends(X):
	assert_raises(ValueError, delete, X, start=-1, end=5)
	assert_raises(ValueError, delete, X, start=100, end=5)
	assert_raises(ValueError, delete, X, start=10, end=5)

	assert_raises(ValueError, delete, X, start=0, end=-1)
	assert_raises(ValueError, delete, X, start=-5, end=-10)
	assert_raises(ValueError, delete, X, start=5, end=5)
	assert_raises(ValueError, delete, X, start=5, end=3)
	assert_raises(ValueError, delete, X, start=5, end=100)


###


def test_multisubstitute_str(X):
	motifs = ['CATCAG', 'CATCAG']
	X_substitute = multisubstitute(X, motifs, spacing=0)

	assert X_substitute.shape == X.shape
	assert X_substitute.sum() == X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_substitute.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_substitute, X)

	new_seq = ('CACATCATCTCATCATCTGCTGACTACTCATCAG' + 
		'CATCAGCTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)

	assert X_substitute.shape == new_seq_ohe.shape
	assert X_substitute.sum() == new_seq_ohe.sum()
	assert_array_almost_equal(X_substitute.sum(dim=-1), new_seq_ohe.sum(dim=-1))
	assert_array_almost_equal(X_substitute, new_seq_ohe)


def test_multisubstitute_str_even(X):
	motifs = ['CATCAG', 'CATCAG']
	X_substitute = multisubstitute(X, motifs, spacing=4)

	assert X_substitute.shape == X.shape
	assert X_substitute.sum() == X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_substitute.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_substitute, X)

	new_seq = ('CACATCATCTCATCATCTGCTGACTACATCAGTA' + 
		'GTCATCAGGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)

	assert X_substitute.shape == new_seq_ohe.shape
	assert X_substitute.sum() == new_seq_ohe.sum()
	assert_array_almost_equal(X_substitute.sum(dim=-1), new_seq_ohe.sum(dim=-1))
	assert_array_almost_equal(X_substitute, new_seq_ohe)


def test_multisubstitute_str_odd(X):
	motifs = ['CATCAG', 'CATCAG']
	X_substitute = multisubstitute(X, motifs, spacing=5)

	assert X_substitute.shape == X.shape
	assert X_substitute.sum() == X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_substitute.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_substitute, X)

	new_seq = ('CACATCATCTCATCATCTGCTGACTACATCAGTA' + 
		'GTCCATCAGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)

	assert X_substitute.shape == new_seq_ohe.shape
	assert X_substitute.sum() == new_seq_ohe.sum()
	assert_array_almost_equal(X_substitute.sum(dim=-1), new_seq_ohe.sum(dim=-1))
	assert_array_almost_equal(X_substitute, new_seq_ohe)


def test_multisubstitute_ohe(X):
	motif = one_hot_encode('CATCAG').unsqueeze(0)
	X_substitute = multisubstitute(X, [motif, motif], spacing=0)

	assert X_substitute.shape == X.shape
	assert X_substitute.sum() == X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_substitute.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_substitute, X)

	new_seq = ('CACATCATCTCATCATCTGCTGACTACTCATCAG' + 
		'CATCAGCTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)

	assert X_substitute.shape == new_seq_ohe.shape
	assert X_substitute.sum() == new_seq_ohe.sum()
	assert_array_almost_equal(X_substitute.sum(dim=-1), new_seq_ohe.sum(dim=-1))
	assert_array_almost_equal(X_substitute, new_seq_ohe)


def test_multisubstitute_str_multi_seqs_one_motif():
	X = random_one_hot((4, 4, 8), random_state=0)
	X_substitute = multisubstitute(X, ['A', 'C', 'GT'], spacing=[0, 0])

	assert_raises(AssertionError, assert_array_almost_equal, X, X_substitute)
	assert_array_almost_equal(X_substitute, [
		[[1, 0, 1, 0, 0, 0, 0, 0],
         [0, 0, 0, 1, 0, 0, 0, 0],
         [0, 0, 0, 0, 1, 0, 0, 0],
         [0, 1, 0, 0, 0, 1, 1, 1]],

        [[0, 0, 1, 0, 0, 0, 0, 1],
         [1, 0, 0, 1, 0, 0, 0, 0],
         [0, 0, 0, 0, 1, 0, 1, 0],
         [0, 1, 0, 0, 0, 1, 0, 0]],

        [[1, 1, 1, 0, 0, 0, 0, 0],
         [0, 0, 0, 1, 0, 0, 0, 0],
         [0, 0, 0, 0, 1, 0, 0, 1],
         [0, 0, 0, 0, 0, 1, 1, 0]],

        [[1, 0, 1, 0, 0, 0, 0, 1],
         [0, 1, 0, 1, 0, 0, 1, 0],
         [0, 0, 0, 0, 1, 0, 0, 0],
         [0, 0, 0, 0, 0, 1, 0, 0]]])



def test_multisubstitute_start(X):
	motifs = ['CATC', 'AGCCC']
	X_substitute = multisubstitute(X, motifs, spacing=[0], start=10)

	assert X_substitute.shape == X.shape
	assert X_substitute.sum() == X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_substitute.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_substitute, X)

	new_seq = ('CACATCATCTCATCAGCCCCTGACTACTGACGTA' +
		'GTCTGACTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)
	assert_array_almost_equal(X_substitute, new_seq_ohe)

	motifs = ['CATC', 'AGC', 'CC']
	X_substitute = multisubstitute(X, motifs, spacing=[0, 0], start=10)
	new_seq = ('CACATCATCTCATCAGCCCCTGACTACTGACGTA' +
		'GTCTGACTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)
	assert_array_almost_equal(X_substitute, new_seq_ohe)


	X_substitute = multisubstitute(X, motifs, spacing=[5, 4], start=10)
	assert X_substitute.shape == X.shape
	assert X_substitute.sum() == X.sum()
	assert_raises(AssertionError, assert_array_almost_equal, 
		X_substitute.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_substitute, X)

	new_seq = ('CACATCATCTCATCATCTGAGCACTACCGACGTAG' + 
		'TCTGACTGACTGACTGACTACTGACTGACTGAC')
	new_seq_ohe = one_hot_encode(new_seq).unsqueeze(0)
	assert_array_almost_equal(X_substitute, new_seq_ohe)	


def test_multisubstitute_raises_alphabet(X):
	motif = one_hot_encode('CACCAG', alphabet=['A', 'C', 'G'])
	assert_raises(ValueError, multisubstitute, X, [motif, motif], [0])
	assert_raises(ValueError, multisubstitute, X, [motif, motif], [0], 
		['A', 'C', 'G'])


def test_multisubstitute_raises_length(X):
	assert_raises(ValueError, multisubstitute, X, ['C'*1000, 'C'], [0])
	assert_raises(ValueError, multisubstitute, X, ['C', 
		one_hot_encode('C'*1000)], [0])


def test_multisubstitute_spacing(X):
	motifs = ['CATGG', 'CAGGA']
	assert_raises(ValueError, multisubstitute, X, motifs, [-1])
	assert_raises(ValueError, multisubstitute, X, motifs, [0, 0])
	assert_raises(ValueError, multisubstitute, X, motifs, [])
	assert_raises(ValueError, multisubstitute, X, motifs, [10000])


###


def test_randomize(X):
	X_rand = randomize(X, start=10, end=30, random_state=0)

	assert len(X_rand.shape) == 4
	assert X_rand.shape == (1, 1, 4, X.shape[-1])
	assert X_rand.sum() == X.sum()
	assert (X_rand != X).any()
	assert (X_rand[:, 0, :, :10] == X[:, :, :10]).all()
	assert (X_rand[:, 0, :, 30:] == X[:, :, 30:]).all()

	assert_raises(AssertionError, assert_array_almost_equal, 
		X_rand.sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_rand, X)


def test_randomize_probs(X):
	X_rand = randomize(X, start=10, end=30, probs=[[1.0, 0.0, 0.0, 0.0]], 
		random_state=0)
	assert_array_almost_equal(X_rand[:, 0, :, 10:30].sum(dim=(0, -1)), 
		[20, 0, 0, 0])

	X_rand = randomize(X.repeat(2, 1, 1), start=10, end=30, 
		probs=[[1.0, 0.0, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0]], random_state=0)
	assert_array_almost_equal(X_rand[0, 0, :, 10:30].sum(dim=-1), [20, 0, 0, 0])
	assert_array_almost_equal(X_rand[1, 0, :, 10:30].sum(dim=-1), [0, 0, 20, 0])


def test_randomize_n(X):
	X_rand = randomize(X, start=10, end=30, n=12, random_state=0)
	assert X_rand.shape == (1, 12, 4, X.shape[-1])
	assert (X_rand[0, 0:1] != X_rand[0]).any()

	X_rand = randomize(X, start=10, end=30, probs=[[1.0, 0.0, 0.0, 0.0]], n=12, 
		random_state=0)
	assert X_rand.shape == (1, 12, 4, X.shape[-1])
	assert (X_rand[0, 0:1] == X_rand[0]).all()

	X_rand = randomize(X.repeat(5, 1, 1), start=10, end=30, n=12, 
		random_state=0)
	assert X_rand.shape == (5, 12, 4, X.shape[-1])
	assert (X_rand[:, 0:1] != X_rand).any()


def test_randomize_raises_ends(X):
	assert_raises(ValueError, randomize, X, start=-3, end=10)
	assert_raises(ValueError, randomize, X, start=500, end=10)
	assert_raises(ValueError, randomize, X, start=5, end=3)
	assert_raises(ValueError, randomize, X, start=5, end=1000)


def test_randomize_raises_probs(X):
	assert_raises(ValueError, randomize, X, start=5, end=10, 
		probs=[[0.1, 0.8, 0.1]])
	assert_raises(ValueError, randomize, X, start=5, end=10, 
		probs=[[0.1, 0.8, 0.1, 0.0], [0.4, 0.3, 0.2]])
	assert_raises(ValueError, randomize, X, start=5, end=10, 
		probs=[0.1, 0.8, 0.1, 0.0])
	assert_raises(ValueError, randomize, X, start=5, end=10, 
		probs=[[0.1, 100.8, 0.1]])


###


def test_shuffle(X):
	X_shuf = shuffle(X, start=10, end=30, random_state=0)

	assert X_shuf.shape == (1, 1, 4, X.shape[-1])
	assert X_shuf.sum() == X.sum()
	assert (X_shuf != X).any()
	assert (X_shuf[:, 0, :, :10] == X[:, :, :10]).all()
	assert (X_shuf[:, 0, :, 30:] == X[:, :, 30:]).all()

	assert_array_almost_equal(X_shuf[:, 0].sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_shuf[:, 0], X)


def test_shuffle_default(X):
	X_shuf = shuffle(X, random_state=0)

	assert X_shuf.shape == (1, 1, 4, X.shape[-1])
	assert X_shuf.sum() == X.sum()
	assert ((X_shuf[:, 0] != X).sum(dim=(-1, -2)) > 0).all()

	assert_array_almost_equal(X_shuf[:, 0].sum(dim=-1), X.sum(dim=-1))
	assert_raises(AssertionError, assert_array_almost_equal, X_shuf[:, 0], X)


def test_shuffle_raises_ends(X):
	assert_raises(ValueError, shuffle, X, start=-3, end=10)
	assert_raises(ValueError, shuffle, X, start=500, end=10)
	assert_raises(ValueError, shuffle, X, start=5, end=3)
	assert_raises(ValueError, shuffle, X, start=5, end=1000)


###


def test_dinucleotide_shuffle():
	motif = one_hot_encode('CATCACGCATACG').unsqueeze(0)
	dimotif = dinucleotide_shuffle(motif, random_state=0)
	assert dimotif.shape == (1, 20, 4, 13)
	assert dimotif.dtype == torch.int8
	assert characters(dimotif[0, 0]) == 'CGCATCACATACG'
	assert characters(dimotif[0, 1]) == 'CATCGCATACACG'

	dimotif = dinucleotide_shuffle(motif, n=5, random_state=0)
	assert dimotif.shape == (1, 5, 4, 13)

	dimotif = dinucleotide_shuffle(torch.cat([motif]*8), n=7, random_state=0)
	assert dimotif.shape == (8, 7, 4, 13)


def test_dinucleotide_shuffle_composition():
	X = random_one_hot((8, 4, 30), random_state=0)
	X_shuf = dinucleotide_shuffle(X, random_state=0)

	for i in range(X.shape[0]):
		seq = characters(X[i])

		dinucs = collections.defaultdict(int)
		for j in range(len(seq)-1):
			dinucs[seq[j:j+2]] += 1

		for j in range(20):
			dinucs_shuffled = collections.defaultdict(int)
			dinucs_seq = characters(X_shuf[i, j])

			for k in range(len(seq)-1):
				dinucs_shuffled[dinucs_seq[k:k+2]] += 1

			for key, value in dinucs_shuffled.items():
				assert dinucs[key] == value 


def test_dinucleotide_shuffle_start_end():
	X = random_one_hot((8, 4, 50), random_state=0)
	X_shuf = dinucleotide_shuffle(X, start=10, end=30, random_state=0)

	assert (X[:, :, :10].unsqueeze(1) == X_shuf[:, :, :, :10]).all()
	assert (X[:, :, 30:].unsqueeze(1) == X_shuf[:, :, :, 30:]).all()
	assert (X[:, :, 10:30].unsqueeze(1) != X_shuf[:, :, :, 10:30]).any()

	for i in range(X.shape[0]):
		seq = characters(X[i, :, 10:30])

		dinucs = collections.defaultdict(int)
		for j in range(len(seq)-1):
			dinucs[seq[j:j+2]] += 1

		for j in range(20):
			dinucs_shuffled = collections.defaultdict(int)
			dinucs_seq = characters(X_shuf[i, j, :, 10:30])

			for k in range(len(seq)-1):
				dinucs_shuffled[dinucs_seq[k:k+2]] += 1

			for key, value in dinucs_shuffled.items():
				assert dinucs[key] == value 	


def test_dinucleotide_shuffle_large_alphabet():
	alpha = ['A', 'C', 'G', 'T', 'N']
	seq = 'CGATCAGCANNCACATCAGCATANNAAT'
	motif = one_hot_encode(seq, alphabet=alpha, ignore=[]).unsqueeze(0)
	dimotif = dinucleotide_shuffle(motif, random_state=0)

	dinucs = collections.defaultdict(int)
	for i in range(len(seq)-1):
		dinucs[seq[i:i+2]] += 1

	for j in range(20):
		dinucs_shuffled = collections.defaultdict(int)
		dinucs_seq = characters(dimotif[0, j], alphabet=alpha)

		for i in range(len(seq)-1):
			dinucs_shuffled[dinucs_seq[i:i+2]] += 1

		for key, value in dinucs_shuffled.items():
			assert dinucs[key] == value 


def test_dinucleotide_shuffle_missing_alphabet():
	seq = 'ATATATTAAAATTATTATATATTTATATATTTAAAAATTTTTAATA'
	motif = one_hot_encode(seq).unsqueeze(0)
	dimotif = dinucleotide_shuffle(motif, random_state=0)

	dinucs = collections.defaultdict(int)
	for i in range(len(seq)-1):
		dinucs[seq[i:i+2]] += 1

	for j in range(20):
		dinucs_shuffled = collections.defaultdict(int)
		dinucs_seq = characters(dimotif[0, j])

		for i in range(len(dinucs_seq)-1):
			dinucs_shuffled[dinucs_seq[i:i+2]] += 1

		for key, value in dinucs_shuffled.items():
			assert dinucs[key] == value 


def test_dinucleotide_shuffle_raises_short():
	motif = one_hot_encode('AATA').unsqueeze(0)
	assert_raises(ValueError, dinucleotide_shuffle, motif)


def test_dinucleotide_shuffle_raises_ohe():
	seq = 'ATATATTAAAATTATTATATATTTATATATTTAAAAATTTTTAATA'
	motif = one_hot_encode(seq).unsqueeze(0) 

	assert_raises(ValueError, dinucleotide_shuffle, motif + 1)
	assert_raises(ValueError, dinucleotide_shuffle, motif.unsqueeze(0))
	assert_raises(ValueError, dinucleotide_shuffle, motif[0])
	assert_raises(ValueError, dinucleotide_shuffle, "ACGTCACGATC")


def test_dinucleotide_shuffle_raises_N():
	seq = 'ATATATTAAAATNNNATTTAAANNNTTTTTAATA'
	motif = one_hot_encode(seq).unsqueeze(0)
	assert_raises(ValueError, dinucleotide_shuffle, motif)


def test_dinucleotide_shuffle_homopolymer():
	seq_ohe = one_hot_encode('AAAAAAAAAAAAAAAAAAAAAAAAAAAAAA').unsqueeze(0)
	assert_raises(ValueError, dinucleotide_shuffle, seq_ohe)



# attribute.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import torch
import torch.nn.functional as F

import warnings

from tqdm import trange
from .ersatz import dinucleotide_shuffle


def hypothetical_attributions(multipliers, X, references):
	"""A function for aggregating contributions into hypothetical attributions.

	When handling categorical data, like one-hot encodings, the gradients
	returned by a method like DeepLIFT/SHAP may need to be modified because
	the choice of one character at a position explicitly means that the other
	characters are not there. So, one needs to account for each character change 
	actually being the addition of one character AND the subtraction of another 
	character. Basically, once you've calculated the multipliers, you need to 
	subtract out the contribution of the nucleotide actually present and then 
	add in the contribution of the nucleotide you are becomming.

	Each element in the tensor is considered an independent example 

	As an implementation note: to be compatible with Captum, each input must
	be a tuple of length 1 and the returned value will be a tuple of length 1.
	I know this sounds silly but it's the most convenient implementation choice 
	to make the function compatible across DeepLiftShap implementations.


	Parameters
	----------
	multipliers: tuple of one torch.tensor, shape=(n_baselines, 4, length)
		The multipliers/gradient calculated by a method like DeepLIFT/SHAP.
		These should include values for both the observed characters and the
		unobserved characters at each position

	X: tuple of one torch.tensor, shape=(n_baselines, 4, length)
		The one-hot encoded sequence being explained

	references: tuple of one torch.tensor, shape=(n_baselines, 4, length)
		The one-hot encoded reference sequences, usually a shuffled version
		of the corresponding sequence in X.


	Returns
	-------
	projected_contribs: tuple of one torch.tensor, shape=(1, 4, length)
		The attribution values for each nucleotide in the input.
	"""

	for val in multipliers, X, references:
		if not isinstance(val, tuple) or len(val) != 1:
			raise ValueError("All inputs must be one-element tuples.")

		if not isinstance(val[0], torch.Tensor):
			raise ValueError("The first element of each input must be a "
				"tensor.")

		if val[0].shape != multipliers[0].shape:
			raise ValueError("Shape of all tensors must match.") 


	projected_contribs = torch.zeros_like(references[0], dtype=X[0].dtype, 
		device=X[0].device)
	
	for i in range(X[0].shape[1]):
		hypothetical_input = torch.zeros_like(X[0], dtype=X[0].dtype, 
			device=X[0].device)
		hypothetical_input[:, i] = 1.0
		hypothetical_diffs = hypothetical_input - references[0]
		hypothetical_contribs = hypothetical_diffs * multipliers[0]

		projected_contribs[:, i] = torch.sum(hypothetical_contribs, dim=1)

	return (projected_contribs,)


def _register_hooks(module): 
	if len(module._backward_hooks) > 0:
		return
	if not isinstance(module, tuple(module._NON_LINEAR_OPS.keys())):
		return

	module.handles = []
	module.handles.append(module.register_forward_hook(_f_hook))
	module.handles.append(module.register_forward_pre_hook(_fp_hook))
	module.handles.append(module.register_full_backward_hook(_b_hook))


def _clear_hooks(module):
	if hasattr(module, "handles") and len(module.handles) > 0:
		for handle in module.handles:
			handle.remove()

		del module.handles


def _fp_hook(module, inputs): 
	module.input = inputs[0].clone().detach()


def _f_hook(module, inputs, outputs):
	module.output = outputs.clone().detach()


def _b_hook(module, grad_input, grad_output):
	return module._NON_LINEAR_OPS[type(module)](module, grad_input, 
		grad_output)


def _nonlinear(module, grad_input, grad_output):
	"""An internal function implementing a general-purpose nonlinear correction.

	This function, copied and slightly modified from Captum, is meant to be
	the `rescale` rule applied to general non-linear functions such as
	activations.
	"""

	delta_in_ = torch.sub(*module.input.chunk(2))
	delta_out_ = torch.sub(*module.output.chunk(2))

	delta_in = torch.cat([delta_in_, delta_in_])
	delta_out = torch.cat([delta_out_, delta_out_])

	delta = delta_out / delta_in
	idxs = torch.abs(delta_in) < 1e-6

	return (torch.where(idxs, grad_input[0], grad_output[0] * delta),)


def _softmax(module, grad_input, grad_output):
	"""An internal function implementing a correction for softmax activations.

	This function, copied and slightly modified from Captum, is meant to be
	the `rescale` rule applied specifically to softmax activations without
	needing to remove them and operate on the underlying logits.
	"""

	delta_in_ = torch.sub(*module.input.chunk(2))
	delta_out_ = torch.sub(*module.output.chunk(2))

	delta_in = torch.cat([delta_in_, delta_in_])
	delta_out = torch.cat([delta_out_, delta_out_])

	delta = delta_out / delta_in
	idxs = torch.abs(delta_in) < 1e-6

	grad_input_unnorm = torch.where(idxs, grad_input[0], grad_output[0] * delta)

	n = grad_input[0].numel()
	new_grad_inp = grad_input_unnorm - grad_input_unnorm.sum() * 1 / n
	return (new_grad_inp,)


def _maxpool(module, grad_input, grad_output):
	"""An internal function implementing a 1D max-pooling correction.

	This function, copied and slightly modified from Captum, is meant to be
	the `rescale` rule applied to max pooling layers given their nature of
	aggregating values across multiple positions.
	"""

	if isinstance(module, torch.nn.MaxPool1d):
		pool_func, unpool_func = F.max_pool1d, F.max_unpool1d
	elif isinstance(module, torch.nn.MaxPool2d):
		pool_func, unpool_func = F.max_pool2d, F.max_unpool2d
	else:
		raise ValueError("module must be either MaxPool1d or MaxPool2d")


	with torch.no_grad():
		delta_in_ = torch.sub(*module.input.chunk(2))
		delta_in = torch.cat([delta_in_, delta_in_])

		output, output_ref = module.output.chunk(2)
		delta_out_xmax = torch.max(output, output_ref)
		delta_out = torch.cat([delta_out_xmax - output_ref, 
			output - delta_out_xmax])

		_, indices = pool_func(module.input, module.kernel_size, module.stride, 
			module.padding, module.dilation, module.ceil_mode, True)

		unpool_ = unpool_func(grad_output[0] * delta_out, indices, 
			module.kernel_size, module.stride, module.padding, 
			list(module.input.shape))
		unpool_delta, unpool_ref_delta = torch.chunk(unpool_, 2)

	unpool_delta_ = unpool_delta + unpool_ref_delta
	unpool_delta = torch.cat([unpool_delta_, unpool_delta_])
	idxs = torch.abs(delta_in) < 1e-7

	new_grad_inp = torch.where(idxs, grad_input[0], unpool_delta / delta_in)
	return (new_grad_inp,)


def deep_lift_shap(model, X, args=None, target=0,  batch_size=32,
	references=dinucleotide_shuffle, n_shuffles=20, return_references=False, 
	hypothetical=False, warning_threshold=0.001, additional_nonlinear_ops=None,
	print_convergence_deltas=False, raw_outputs=False, device='cuda', 
	random_state=None, verbose=False):
	"""Calculate attributions for a set of sequences using DeepLIFT/SHAP.

	This function will calculate the DeepLIFT/SHAP attributions on a set of
	sequences given a model. These attributions have the additive property that
	the sum of the attributions is ~equal to the difference in prediction
	between the original sequence and the reference sequences.

	As an implementation note, the batch size refers to the number of
	example-reference pairs that are being run simultaneously. When the batch
	size is smaller than the number of references, multiple batches will be
	run per example and the attributions will only be averaged only the
	references after they have all been covered. You may want to do this if the
	model or examples are so large that only a few can fit in memory at a time.
	The result will be identical to if all examples could fit in memory and
	each batch contained all the references.

	Convergence deltas are calculated automatically for each example-reference
	pair. Theoretically, these should be zero, but may in practice just be a
	small number due to machine precision issues with non-linear models. If
	these deltas exceed a warning threshold, a non-terminating warning will be 
	issued to let you know that the deltas have been exceeded.

	NOTE: predictions MUST yield a `(batch_size, n_targets)` tensor, even if
	n_targets is 1. If your model yields something more complicated you must
	wrap the model in a small class that operates on the outputs in a manner
	that yields such a tensor, e.g., by slicing the output or summing along
	a relevant axis.


	Parameters
	----------
	model: torch.nn.Module
		A PyTorch model to use for making predictions. These models can take in
		any number of inputs and make any number of outputs. The additional
		inputs must be specified in the `args` parameter.

	X: torch.tensor, shape=(-1, len(alphabet), length)
		A set of one-hot encoded sequences to calculate attribution values
		for. 

	args: tuple or None, optional
		An optional set of additional arguments to pass into the model. If
		provided, each element in the tuple or list is one input to the model
		and the element must be formatted to be the same batch size as `X`. If
		None, no additional arguments are passed into the forward function.
		Default is None.

	target: int, optional
		The output of the model to calculate gradients/attributions for. This
		will index the last dimension of the predictions. Default is 0.

	batch_size: int, optional
		The number of sequence-reference pairs to pass through DeepLiftShap at
		a time. Importantly, this is not the number of elements in `X` that
		are processed simultaneously (alongside ALL their references) but the
		total number of `X`-`reference` pairs that are processed. This means
		that if you are in a memory-limited setting where you cannot process
		all references for even a single sequence simultaneously that the
		work is broken down into doing only a few references at a time. Default
		is 32.

	references: func or torch.Tensor, optional
		If a function is passed in, this function is applied to each sequence
		with the provided random state and number of shuffles. This function
		should serve to transform a sequence into some form of signal-null
		background, such as by shuffling it. If a torch.Tensor is passed in,
		that tensor must have shape `(len(X), n_shuffles, *X.shape[1:])`, in
		that for each sequence a number of shuffles are provided. Default is
		the function `dinucleotide_shuffle`. 

	n_shuffles: int, optional
		The number of shuffles to use if a function is given for `references`.
		If a torch.Tensor is provided, this number is ignored. Default is 20.

	return_references: bool, optional
		Whether to return the references that were generated during this
		process. Only use if `references` is not a torch.Tensor. Default is 
		False. 

	hypothetical: bool, optional
		Whether to return attributions for all possible characters at each
		position or only for the character that is actually at the sequence.
		Practically, whether to return the returned attributions from captum
		with the one-hot encoded sequence. Default is False.

	warning_threshold: float, optional
		A threshold on the convergence delta that will always raise a warning
		if the delta is larger than it. Normal deltas are in the range of
		1e-6 to 1e-8. Note that convergence deltas are calculated on the
		gradients prior to the aggr_func being applied to them. Default 
		is 0.001. 

	additional_nonlinear_ops: dict or None, optional
		If additional nonlinear ops need to be added to the dictionary of
		operations that can be handled by DeepLIFT/SHAP, pass a dictionary here
		where the keys are class types and the values are the name of the
		function that handle that sort of class. Make sure that the signature
		matches those of `_nonlinear` and `_maxpool` above. This can also be
		used to overwrite the hard-coded operations by passing in a dictionary
		with overlapping key names. If None, do not add any additional 
		operations. Default is None.

	print_convergence_deltas: bool, optional
		Whether to print the convergence deltas for each example when using
		DeepLiftShap. Default is False.

	raw_outputs: bool, optional
		Whether to return the raw outputs from the method -- in this case,
		the multipliers for each example-reference pair -- or the processed
		attribution values. Default is False.

	device: str or torch.device, optional
		The device to move the model and batches to when making predictions. If
		set to 'cuda' without a GPU, this function will crash and must be set
		to 'cpu'. Default is 'cuda'. 

	random_state: int or None or numpy.random.RandomState, optional
		The random seed to use to ensure determinism. If None, the
		process is not deterministic. Default is None. 

	verbose: bool, optional
		Whether to display a progress bar. Default is False.


	Returns
	-------
	attributions: torch.tensor
		If `raw_outputs=False` (default), the attribution values with shape
		equal to `X`. If `raw_outputs=True`, the multipliers for each example-
		reference pair with shape equal to `(X.shape[0], n_shuffles, X.shape[1],
		X.shape[2])`. 

	references: torch.tensor, optional
		The references used for each input sequence, with the shape
		(n_input_sequences, n_shuffles, 4, length). Only returned if
		`return_references = True`. 
	"""

	_NON_LINEAR_OPS = {
		torch.nn.ReLU: _nonlinear,
		torch.nn.ReLU6: _nonlinear,
		torch.nn.RReLU: _nonlinear,
		torch.nn.SELU: _nonlinear,
		torch.nn.CELU: _nonlinear,
		torch.nn.GELU: _nonlinear,
		torch.nn.SiLU: _nonlinear,
		torch.nn.Mish: _nonlinear,
		torch.nn.GLU: _nonlinear,
		torch.nn.ELU: _nonlinear,
		torch.nn.LeakyReLU: _nonlinear,
		torch.nn.Sigmoid: _nonlinear,
		torch.nn.Tanh: _nonlinear,
		torch.nn.Softplus: _nonlinear,
		torch.nn.Softshrink: _nonlinear,
		torch.nn.LogSigmoid: _nonlinear,
		torch.nn.PReLU: _nonlinear,
		torch.nn.MaxPool1d: _maxpool,
		torch.nn.MaxPool2d: _maxpool,
		torch.nn.Softmax: _softmax
	}
	
	# Misc. set up for overriding operations

	if additional_nonlinear_ops is not None:
		for key, value in additional_nonlinear_ops.items():
			_NON_LINEAR_OPS[key] = value

	model = model.to(device).eval()
	for module in model.modules():
		module._NON_LINEAR_OPS = _NON_LINEAR_OPS

	try:
		model.apply(_register_hooks)
	except Exception as e:
		model.apply(_clear_hooks)
		raise(e)
	
	# Begin DeepLIFT procedure
	
	attributions, references_, Xi, rj, attr_ = [], [], [], [], []
	if isinstance(references, torch.Tensor):
		n_shuffles = references.shape[1]
	n, z = X.shape[0] * n_shuffles, 0

	for i in trange(n, disable=not verbose):
		Xi.append(i // n_shuffles)
		rj.append(i % n_shuffles)

		if len(Xi) == batch_size or i == (n-1):
			_X = X[Xi].cpu()
			_args = None if args is None else tuple([a[Xi].to(device) 
				for a in args])

			# Handle reference sequences while ensuring that the same seed is
			# used for each shuffle even if not all shuffles are done in the
			# same batch.
			if isinstance(references, torch.Tensor):
				_references = references[Xi, rj]
			else:
				if random_state is None:
					_references = references(_X, n=1)[:, 0]
				else:
					_references = torch.cat([references(_X[j:j+1], n=1, 
						random_state=random_state+rj[j])[:, 0] 
							for j in range(len(_X))])

			_X = _X.to(device).requires_grad_()
			_references = _references.to(device).requires_grad_()

			# This next block is actually running DeepLIFT by concatenating the
			# batch of examples and the batch of references and running the
			# forward and backward passes that have been modified by the above
			# hooks. In a try-except block to make sure we remove hooks if an
			# error is raised. 
			try:
				X_ = torch.cat([_X, _references])

				# Calculate the gradients using the rescale rule
				with torch.autograd.set_grad_enabled(True):
					if _args is not None:
						_args = (torch.cat([arg, arg]) for arg in _args)
						y = model(X_, *_args)[:, target]
					else:
						y = model(X_)[:, target]

					multipliers = torch.autograd.grad(y.sum(), _X)[0]

				# Check that the prediction-difference-from-reference is equal to
				# the sum of the attributions
				output_diff = torch.sub(*torch.chunk(y, 2))
				input_diff = torch.sum((_X - _references) * multipliers, 
					dim=(1, 2))
				convergence_deltas = abs(output_diff - input_diff)

				if torch.any(convergence_deltas > warning_threshold):
					warnings.warn("Convergence deltas too high: " +   
						str(convergence_deltas), RuntimeWarning)

				if print_convergence_deltas:
					print(convergence_deltas)

			except Exception as e:
				model.apply(_clear_hooks)
				raise(e)

			# If not returning the raw multipliers then apply the correction for
			# character encodings
			if raw_outputs == False:
				multipliers = hypothetical_attributions((multipliers,), (_X,), 
					(_references,))[0]

			# attr_ is a list where each element is a tensor for the multipliers
			# of one example so that we can chunk them together once all
			# references for an example are 
			attr_.extend(list(multipliers.cpu().detach()))

			# When all references for a sequence have been calculated, remove
			# that block from the list of example-reference attributions and
			# add it to the final attribution list, averaging across references
			# if providing the processed results.
			while len(attr_) >= n_shuffles:
				attr_chunk = torch.stack(attr_[:n_shuffles])

				if raw_outputs == False:
					attr_chunk = attr_chunk.mean(dim=0)
					if not hypothetical:
						attr_chunk *= X[z].cpu()

				attributions.append(attr_chunk)
				attr_ = attr_[n_shuffles:]
				z += 1

			if return_references:
				references_.extend(list(_references.cpu().detach()))

			Xi, rj = [], []


	model.apply(_clear_hooks)
	for module in model.modules():
		del(module._NON_LINEAR_OPS)

	attributions = torch.stack(attributions)

	if return_references:
		references_ = torch.cat(references_).reshape(X.shape[0], n_shuffles, 
			*X.shape[1:])
		return attributions, references_
	return attributions


def _captum_deep_lift_shap(model, X, args=None, target=0, batch_size=32,
	references=dinucleotide_shuffle, n_shuffles=20,  return_references=False, 
	hypothetical=False, device='cuda', random_state=None, verbose=False):
	"""Calculate attributions using DeepLift/Shap and a given model. 

	This function will calculate DeepLift/Shap attributions on a set of
	sequences. It assumes that the model returns "logits" in the first output,
	not softmax probabilities, and count predictions in the second output.
	It will create GC-matched negatives to use as a reference and proceed
	using the given batch size.

	This is an internal/debugging function that is mostly meant to be used to
	check for differences with the `deep_lift_shap` method.


	Parameters
	----------
	model: torch.nn.Module
		A PyTorch model to use for making predictions. These models can take in
		any number of inputs and make any number of outputs. The additional
		inputs must be specified in the `args` parameter.

	X: torch.tensor, shape=(-1, len(alphabet), length)
		A set of one-hot encoded sequences to calculate attribution values
		for. 

	args: tuple or None, optional
		An optional set of additional arguments to pass into the model. If
		provided, each element in the tuple or list is one input to the model
		and the element must be formatted to be the same batch size as `X`. If
		None, no additional arguments are passed into the forward function.
		Default is None.

	target: int, optional
		The output of the model to calculate gradients/attributions for. This
		will index the last dimension of the predictions. Default is 0.

	batch_size: int, optional
		The number of sequence-reference pairs to pass through DeepLiftShap at
		a time. Importantly, this is not the number of elements in `X` that
		are processed simultaneously (alongside ALL their references) but the
		total number of `X`-`reference` pairs that are processed. This means
		that if you are in a memory-limited setting where you cannot process
		all references for even a single sequence simultaneously that the
		work is broken down into doing only a few references at a time. Default
		is 32.

	references: func or torch.Tensor, optional
		If a function is passed in, this function is applied to each sequence
		with the provided random state and number of shuffles. This function
		should serve to transform a sequence into some form of signal-null
		background, such as by shuffling it. If a torch.Tensor is passed in,
		that tensor must have shape `(len(X), n_shuffles, *X.shape[1:])`, in
		that for each sequence a number of shuffles are provided. Default is
		the function `dinucleotide_shuffle`. 

	n_shuffles: int, optional
		The number of shuffles to use if a function is given for `references`.
		If a torch.Tensor is provided, this number is ignored. Default is 20.

	return_references: bool, optional
		Whether to return the references that were generated during this
		process. Only use if `references` is not a torch.Tensor. Default is 
		False. 

	hypothetical: bool, optional
		Whether to return attributions for all possible characters at each
		position or only for the character that is actually at the sequence.
		Practically, whether to return the returned attributions from captum
		with the one-hot encoded sequence. Default is False.

	device: str or torch.device, optional
		The device to move the model and batches to when making predictions. If
		set to 'cuda' without a GPU, this function will crash and must be set
		to 'cpu'. Default is 'cuda'. 

	random_state: int or None or numpy.random.RandomState, optional
		The random seed to use to ensure determinism. If None, the
		process is not deterministic. Default is None. 

	verbose: bool, optional
		Whether to display a progress bar. Default is False.


	Returns
	-------
	attributions: torch.tensor
		The attributions calculated for each input sequence, with the same
		shape as the input sequences.

	references: torch.tensor, optional
		The references used for each input sequence, with the shape
		(n_input_sequences, n_shuffles, 4, length). Only returned if
		`return_references = True`. 
	"""

	from captum.attr import DeepLiftShap as CaptumDeepLiftShap

	model = model.to(device).eval()

	attributions = []
	references_ = []
	with torch.no_grad():
		for i in trange(len(X), disable=not verbose):
			_X = X[i:i+1].to(device)
			_args = None if args is None else tuple([a[i:i+1].to(device) 
				for a in args])

			# Calculate references
			if isinstance(references, torch.Tensor):
				_references = references[i].to(device)
			else:
				_references = references(_X, n=n_shuffles, 
					random_state=random_state)[0].to(device)
						
			attr = CaptumDeepLiftShap(model).attribute(_X, _references, 
				target=target, additional_forward_args=_args, 
				custom_attribution_func=hypothetical_attributions)

			if not hypothetical:
				attr = (attr * _X)
			
			if return_references:
				references_.append(_reference.unsqueeze(0))

			attributions.append(attr.cpu())

	attributions = torch.cat(attributions)

	if return_references:
		return attributions, torch.cat(references_)
	return attributions